"""

"""

import time
import functools
import logging
import sys
import random
import traceback
import json
import gc
import pandas as pd
import numpy as np

# 导入常用的固定路径(多平台通用)
from kw618._file_path import *
from kw618.k_pandas.utils_pandas import *

# AES加密使用
import base64
#注：python3 安装 Crypto 是 pip3 install -i https://pypi.tuna.tsinghua.edu.cn/simple pycryptodome<br><br>
from Crypto.Cipher import AES
import hashlib

# 个人用于记录报错内容的log装饰器
def log_error(log_directory=f"{FILE_PATH_FOR_ZIRU_CODE}/Log/ttt_log", throw_error=False):
    # 作为装饰器时, 一定要加上(); 否则就不会返回内部的decorate函数了
    # 如果没有传入log的存放目录, 默认使用上述目录
    def decorate(func):
        def record_error(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                module_name = get_this_module_name()
                func_name = func.__name__ # 暂时没利用, 可删
                kkprint(module_name=module_name, func_name=func_name)
                tb_txt = traceback.format_exc(limit=5) # limit参数: 表示traceback最多到第几层
                log_file_path = f"{log_directory}/{module_name}_error.log"
                with open(log_file_path, "a", encoding="utf-8") as f:
                    print(f"\n【捕获到异常】\n{tb_txt}\n【异常存储路径】: {log_file_path}\n")
                    log_msg = tb_txt
                    this_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                    f.write(f"{this_time}\n{log_msg}\n\n\n")
                # 有时候需要把错误内容抛出, 在更外层捕获 (通过'消费者多线程池'来捕获,让其url进入到error_queue)
                if throw_error:
                    raise Exception(tb_txt)
        return record_error
    return decorate



# python官网的例子
def logged(level, name=None, message=None):
    """
    这是python cookbook 中官方写写的log案例
    Add logging to a function. level is the logging
    level, name is the logger name, and message is the
    log message. If name and message aren't specified,
    they default to the function's module and name.

    可以看到, 如果你想要给装饰器传参, 就需要在decorate外面再嵌套一层函数: 总共3层
    """
    def decorate(func): # 此处一定只有一个func形参
        logname = name if name else func.__module__
        log = logging.getLogger(logname)
        logmsg = message if message else func.__name__

        @functools.wraps(func) # 这里的装饰器可以修改__name__的问题(其实没啥用, 反正写上更好就对了, 管他呢)
        def wrapper(*args, **kwargs):  # 此处的形参一定是(*args, **kwargs), 并且与下面return中传入的参数一致!!
            log.log(level, logmsg)
            return func(*args, **kwargs) # 一定要记得return
        return wrapper  # 返回的函数名称一定和上面定义(warpper)的一致!!
    return decorate

# Example use
# @logged(logging.DEBUG)
# def add(x, y):
#     return x + y
#
# @logged(logging.CRITICAL, 'example')
# def spam():
#     print('Spam!')




def timer(func):
    """装饰器：记录并打印函数耗时"""
    def decorated(*args, **kwargs):
        st = time.time()
        ret = func(*args, **kwargs)
        print('执行时长: {} 秒'.format(time.time() - st))
        return ret
    return decorated



def get_this_module_name():
    "获取本函数所在脚本的模块名称"
    argv_str = sys.argv[-1]
    return argv_str.split("/")[-1][:-3]


def sprint(**kwargs):
    """
        主要使用场景: 调试bug时候, 经常要打印某个变量名, 是否正确得到, 每次手动写print, 烦的一批!!!
                    遂, 写成了通用接口
        tips:
            与kprint的区别:  sprint只显示一行, 更加简洁   (意为: simply print)
    """
    "kwargs就是一个dict类型"
    for k, v in kwargs.items():
        print(f"\n变量'{k}'--({type(v)})----> {v}\n")

def kprint(**kwargs):
    """
        主要使用场景: 调试bug时候, 经常要打印某个变量名, 是否正确得到, 每次手动写print, 烦的一批!!!
                    遂, 写成了通用接口
    """
    "kwargs就是一个dict类型"
    for k, v in kwargs.items():
        # print(f"\n【变量'{k}'】: {v}, {type(v)}\n\n")

        # int型是没有len()方法的, 所以做了一个判断....(麻烦)
        if hasattr(v, "__len__"): # 是否有__len__魔法方法 (用于测算长度用的)
            print(f"\n【变量'{k}'】:\n    类型: {type(v)}\n    长度: {len(v)}\n    值: {v} \n\n")
        else:
            print(f"\n【变量'{k}'】:\n    类型: {type(v)}\n    长度: NAN\n    值: {v} \n\n")


def kkprint(**kwargs):
    """
        主要使用场景: 当爬虫接口中的json太长时候, 为了美化打印用的.
    """
    "方便打印出某些变量的值(测试使用); 需要使用关键字传参"
    json_ = json.dumps(kwargs, indent=4, ensure_ascii=False)
    print(json_)





def k_update(dic, key, value):
    "添加一个'k-v'对的同时, 返回这个添加后的dict对象!! (python默认是没有返回值的, 有些时候不方便) [下同]"
    dic[str(key)] = value
    return dic

def k_append(lst, element):
    lst.append(element)
    return lst

def k_extend(lst, lst2):
    lst.extend(lst2)
    return lst


def k_memory(obj, accuracy=False):
    """
        params:
            obj: 需要计算内存大小的对象
            accuracy: 是否需要精细测算内存大小 (使用递归的方法,把列表/字典中的元素全部遍历出来计算)

        return:
            memory_usage, unit

        note:
            getsizeof函数默认返回bytes(字节/大B)

        tips:
            比 df["<col>"].memory_usage(deep=True) 要稍大一丢丢 (但可以认为是相同的)
    """

    def recur_cal_memory(obj):
        "递归函数: 计算有深度的对象的内存"

        # 1. 列表对象
        if type(obj) == list:
            total_memory = 0
            for e in obj:
                memory_of_e = recur_cal_memory(e)
                total_memory += memory_of_e
            return total_memory

        # 2. 字典对象
        elif type(obj) == dict:
            total_memory = 0
            for k, v in obj.items():
                memory_of_k = sys.getsizeof(k)
                memory_of_v = recur_cal_memory(v)
                total_memory = total_memory + memory_of_k + memory_of_v
            return total_memory

        # 3. 其他
        else:
            # [递归的'原子条件': 当数据类型不为"list"和"dict"时]
            memory_usage = sys.getsizeof(obj)
            return memory_usage



    # 1. 需要精准计算
    if accuracy:
        memory_usage = recur_cal_memory(obj)

    # 2. 粗略计算即可
    else:
        memory_usage = sys.getsizeof(obj)

    # 以一种更"human"的方式呈现 '内存大小'
    if memory_usage < 1024:
        return round(memory_usage, 2), "Bytes"
    elif memory_usage < 1024*1024:
        return round(memory_usage/1024, 2), "KB"
    elif memory_usage < 1024*1024*1024:
        return round(memory_usage/1024/1024, 2), "MB"
    elif memory_usage < 1024*1024*1024*1024:
        return round(memory_usage/1024/1024/1024, 2), "GB"

def get_df_deep_memory(df):
    show_dict = {}
    columns = list(df.columns)
    for col in columns:
        # df[col].memory_usage(deep=True)
        memory_usage_tuple = k_memory(df[col])
        show_dict.update({col:memory_usage_tuple})
    kprint(show_dict=show_dict)
    return show_dict



def get_top(df, field_1, field_2, top_num=5, ascending=True):
    """
        function: 计算"某"个分类的"某"个字段的"前5名"
        params:
            df: 所需df
            field_1: 按它分类
            field_2: 按它排名
            top_num: 取前几名
    """
    # 先对df的 "field_2" 进行排序
    df = df.sort_values(field_2, ascending=ascending)

    # 用于计数的dict
    d = {}
    def foo(row):
        # nonlocal d
        """
            row: 是df中的一行
        """
        _key = row.get(field_1)
        if d.get(_key, 0) == 0:
            d.update({_key : 1})
            return row
        elif d.get(_key) < top_num:
            d.update({_key : d.get(_key) + 1})
            return row

    # 使用apply, 应用上面的函数
    df2 = df.apply(foo, axis=1)
    _ = df2.columns[0]
    df3 = df2.query(f"{_} == {_}")

    output_data(df3, f"top5_{field_2}")

    return df3


def base64_encrypt(data):
    """
    in:
        data: str类型 / bytes类型
    out:
        encrypted_b: bytes类型
    """
    # base64编码
    if type(data) == str:
        encrypted_b = base64.b64encode(data.encode('utf-8'))
    elif type(data) == bytes:
        encrypted_b = base64.b64encode(data)
    print(f"base64加密后的字节码: {encrypted_b}\n")
    return encrypted_b


def base64_decrypt(b):
    """
    in:
        b: bytes类型
    out:
        origin_s: str类型
    """
    # base64解码
    origin_s = base64.b64decode(b).decode("utf-8")
    print(f"base64解密后的'原始字符串': {origin_s}\n")
    return origin_s





class kwEncryption():
    "支持中文的AES加密!!(mode:CBC模式)"

    def __init__(self, key):
        """
        params:
            key: 必须是ascii字符 (不能是中文) (可以不要求16个字符, 因为后续会自动填充)

        function:
            1. 初始化 key值 和 iv值

        notes:
            1. key,iv使用同一个值
            2. key值和iv值必须要16个字节才行, 所以当key小于16位的时候, 使用"!"自动填充
        """
        # key 和 iv 使用同一个值
        key = key.ljust(16,'!')
        self.key = key
        self.iv = key
        self.key_bytes = bytes(key, encoding='utf-8')
        self.iv_bytes = bytes(key, encoding='utf-8')



    # 用于填充16位字节的辅助函数
    # (不太重要, 但必须要有. 其实可以直接写在底层,不需要自己造轮子啊....!!!贼烦)
    def pkcs7padding(self, text):
        """
        明文使用PKCS7填充
        最终调用AES加密方法时，传入的是一个byte数组，要求是16的整数倍，因此需要对明文进行处理
        :param text: 待加密内容(明文)
        :return:
        """
        bs = AES.block_size  # 16
        length = len(text)
        bytes_length = len(bytes(text, encoding='utf-8'))
        # tips：utf-8编码时，英文占1个byte，而中文占3个byte
        padding_size = length if(bytes_length == length) else bytes_length
        padding = bs - padding_size % bs
        # tips：chr(padding)看与其它语言的约定，有的会使用'\0'
        padding_text = chr(padding) * padding
        return text + padding_text
    def pkcs7unpadding(self, text):
        """
        处理使用PKCS7填充过的数据
        :param text: 解密后的字符串
        :return:
        """
        try:
            length = len(text)
            unpadding = ord(text[length-1])
            return text[0:length-unpadding]
        except Exception as e:
            pass


    def aes_encode(self, content):
        """
        function: AES加密
        参数:
            content: 待加密的内容(原内容)
        模式: cbc
        填充: pkcs7
        return:
            加密后的内容

        """
        cipher = AES.new(self.key_bytes, AES.MODE_CBC, self.iv_bytes)
        # 处理明文
        content_padding = self.pkcs7padding(content)
        # 加密
        aes_encode_bytes = cipher.encrypt(bytes(content_padding, encoding='utf-8'))
        # 重新编码
        result = str(base64.b64encode(aes_encode_bytes), encoding='utf-8')
        return result


    def aes_decode(self, content):
        """
        function: AES解密
        参数:
            content: 加密后的内容
        模式: cbc
        去填充: pkcs7
        return:
            加密前的内容(原内容)
        """
        try:
            cipher = AES.new(self.key_bytes, AES.MODE_CBC, self.iv_bytes)
            # base64解码
            aes_encode_bytes = base64.b64decode(content)
            # 解密
            aes_decode_bytes = cipher.decrypt(aes_encode_bytes)
            # 重新编码
            result = str(aes_decode_bytes, encoding='utf-8')
            # 去除填充内容
            result = self.pkcs7unpadding(result)
        except Exception as e:
            raise Exception(f"该内容:'{content}', 无法被AES解密!!\n")
            # pass
        # if result == None:
        #     return ""
        # else:
        #     return result



# 测试:
# ==============
# 对中文加密
x = kwEncryption("kw618").aes_encode("萧山")
# 对中文解密
s = kwEncryption("kw618").aes_decode(x)





def task_01():
    print("\n\n\ntask正在执行....\n\n\n")


class TimerControler():
    def __init__(self, main_freq=8):
        pass
        """
        params:
            task_freq_table: 任务频次表  (记录每个对应时间的task执行次数)
                note:
                    1. 24-8: 包括24点, 也包括8点  (左右都能取到)
                    2. 使用分段的形式, 避免24个时间段的穷举法...(穷举有点low)

        """
        self.task_freq_table = {
            "24-8":0, "9-11":main_freq, "12-13":5, "14-18":main_freq, "19-21":5, "22-23":3
        }



    def get_task_freq(self, this_hour):
        """
        function:
            从原先的对应关系表中, 获取该小时中, task需要执行的次数!
        params:
            task_freq: 一小时中task需要执行的次数
        """
        if this_hour == 24 or this_hour <=8:
            task_freq = self.task_freq_table.get("24-8", 2)

        elif 9 <= this_hour <= 11:
            task_freq = self.task_freq_table.get("9-11", 2)

        elif 12 <= this_hour <= 13:
            task_freq = self.task_freq_table.get("12-13", 2)

        elif 14 <= this_hour <= 18:
            task_freq = self.task_freq_table.get("14-18", 2)

        elif 19 <= this_hour <= 21:
            task_freq = self.task_freq_table.get("19-21", 2)

        elif 22 <= this_hour <= 23:
            task_freq = self.task_freq_table.get("22-23", 2)

        print(f"\ntask_freq: {task_freq}\n")
        return task_freq




    def accumulate_lst(self, lst):
        """
        function: 计算一个list中的累计值
            eg:
                原lst:   [8, 4, 11, 9, 3, 8, 7, 7]
                累计lst:  [8, 12, 23, 32, 35, 43, 50, 57]
        """
        # lst = [8, 4, 11, 9, 3, 8, 7, 7]
        accumulated_lst = []
        for i in range(1, len(lst)+1):
            accumulated_lst.append(sum(lst[:i]))
        print(f"\naccumulated_lst: {accumulated_lst}\n")
        return accumulated_lst


    def get_minute_run_lst(self, task_freq=6, std_deviation=1):
        """
        params:
            std_deviation(标准差): 默认为3; 大概意思就是在均值的基础上上下浮动3
        """
        if task_freq != 0:
            avg_value = 60 / task_freq

            minute_array = np.random.normal(avg_value, std_deviation, task_freq)
            minute_array = [int(min) for min in minute_array] # 取整数
            print(f"\n 60分钟内的随机取数为: {minute_array}\n")
            minute_run_lst = self.accumulate_lst(minute_array)
            print(f"\nminute_run_lst: {minute_run_lst}\n")
            return minute_run_lst # [7, 17, 27, 37, 47, 57] (list)  (当lst中存在>=60的分钟数, 后续会被忽略, 不执行)
        else:
            return [999] # 当这个小时段不需要执行task的时候, 故意把"执行的实际分钟数调至999(最多是59)",即: 该小时不执行



    def run(self, task):
        """
            function:
                1. 根据任务执行频次表"", 定时执行task函数. (模拟人工审批, 以免被识别成是机器爬虫)
                2. run方法为该"时间控制器"的主入口

        """
        init = True
        while True:
            today_obj = pd.to_datetime("today")
            this_time = today_obj.strftime("%X")
            this_hour = int(this_time[0:2])
            this_minute = int(this_time[3:5])
            this_second = int(this_time[6:9])

            # 1. 每整小时执行一次: 重新执行一遍 "抽取随机分钟"
            if (this_minute == 0 and this_second == 0) or (init == True):
                if init == True:
                    init = False # 如果一开始执行的时候没有整点, 就只能通过这种方式产生"minute_run_lst"
                    print(f"初始化报时: {this_time}")
                    print("===========================================")
                    print("===========================================\n\n")
                else:
                    print(f"整小时报时: {this_time} ----------")
                    print("===========================================")
                    print("===========================================\n\n")
                task_freq = self.get_task_freq(this_hour)
                minute_run_lst = self.get_minute_run_lst(task_freq=task_freq)
                minute_run_lst.append(this_minute+1) # (当初始化执行的时候,需要马上执行)
                # 如果是59的话就不执行了, 防止task来不及执行完, 错过下一个整点的"抽取"机会....导致下一个小时段, 都没有新的"minute_run_lst"
                if 59 in minute_run_lst:
                    minute_run_lst.remove(59)
                print(f"\n\n【minute_run】【ALL】当前小时段的minute_run_lst: {minute_run_lst}\n\n")


            # 2. 每整分钟执行一次: 去run_time_lst中检查下, 当前分钟是否需要执行task
            if this_second == 0:
                print(f"\n整分钟报时: {this_time}\n")
                if this_minute in minute_run_lst.copy():
                    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>开始执行任务:")
                    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
                    task() # 执行任务
                    print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<任务执行结束;")
                    print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
                    minute_run_lst.remove(this_minute) # 执行完后, 把这个 minute_run 的时间给删掉 (防止同一分钟的不同秒中又会重复运行)
                    print(f"\n\n【minute_run】【剩余】当前小时段的minute_run_lst: {minute_run_lst}\n\n")

            time.sleep(1)

# 测试:
# ==============
# TC = TimerControler()
# TC.run(task_01)



def is_same_pos_neg(x, y):
    # 用于检测传入的x,y的数据类型
    if hassttr(x, "__neg__") and hassttr(y, "__neg__"):
        if x * y > 0:
            print("x与y, 同正负方向.\n")
            return True
        elif x * y < 0:
            print("x与y, 不同正负方向.\n")
            return False # 此时, 为'买卖点'
        elif x * y == 0:
            print("x与y, 其中有个为0.\n")
            return True # 主要用途是"量化中发生金叉/死叉时候判断买卖点, 为0时不执行买卖, 所以把它暂时视为True"
    else:
        raise(Exception("传入的x/y, 数据类型错误...\n"))





if __name__ == "__main__":
    m = 33
    n = 99
    kprint(m=m, n=n)
    TimerControler().get_minute_run_lst()








#



from kw618.k_requests.utils_requests import *
from kw618.k_pandas.utils_pandas import *

# 导入常用的固定路径(多平台通用)
from kw618._file_path import *

req = myRequest().req

def get_wyy_mp3(save_file_name="", save_folder_path=".", wyy_url="https://music.163.com/song?id=505688656&userid=80581305"):
    id = re.findall(r"id=(\d+)", wyy_url)
    id = id[0] if id else ""
    if id == "":
        k_success = "False"
        k_msg = "找不到id\n"
    else:
        api_url = "https://tenapi.cn/wyy?id=" + id
        print(f"api_url: {api_url}")
        res_obj = req(api_url, is_obj=True)
        res_dict = res_obj.json()
        print(res_dict)
        if res_dict.get("msg") == "解析成功":
            mp3_url = res_dict.get("url", "")
            res_obj = req(mp3_url, is_obj=True, use_session=True)
            res_content = res_obj.content
            mp3_content = res_content
            # 存储mp3音频
            if save_file_name == "":
                save_file_name = f"网易云音乐_{get_sim_today_date()}_{get_sim_this_time()}"
            save_full_path = f"{save_folder_path}/{save_file_name}.mp3"
            with open(save_full_path, "wb") as f:
                f.write(mp3_content)
            k_success = "True"
            k_msg = f"v1: 音乐'{save_file_name}':下载成功"
        else:
            k_success = "False"
            k_msg = "v1: api接口解析id失败"

    response_dict = {
        "k_success" : k_success, "k_msg" : k_msg
    }
    return response_dict



def get_wyy_mp3_v2(save_file_name="", save_folder_path=".", wyy_url="https://music.163.com/song?id=505688656&userid=80581305"):
    """
        这个api的解析能力比上面的强, 基本上上面解析不出来的, 这个都行, 可屌了"
        有时候在python运行这个api也会经常连接失败, 但是到浏览器上解析又是有用的, 很奇怪...
        比如这个:
            https://www.apicp.cn/API/wyy/api.php?id=509252160
            https://www.apicp.cn/API/wyy/api.php?id=436514221
    """

    id = re.findall(r"id=(\d+)", wyy_url)
    id = id[0] if id else ""
    if id == "":
        k_success = "False"
        k_msg = "找不到id\n"
    else:
        api_url = "https://www.apicp.cn/API/wyy/api.php?id=" + id
        print(f"api_url: {api_url}")
        res_obj = req(api_url, is_obj=True, use_session=True)
        res_content = res_obj.content
        if res_content:
            print("v2: 解析成功")
            mp3_content = res_content
            # 存储mp3音频
            if save_file_name == "":
                save_file_name = f"网易云音乐_{get_sim_today_date()}_{get_sim_this_time()}"
            save_full_path = f"{save_folder_path}/{save_file_name}.mp3"
            with open(save_full_path, "wb") as f:
                f.write(mp3_content)
        else:
            print("v2: 解析失败")




def get_mp3(save_file_name="", wyy_url="https://music.163.com/song?id=505688656&userid=80581305"):
    """
        save_file_name: 不能包含"/"等特殊字符!!  (用逗号取代"/")
    """
    save_folder_path = FILE_PATH_FOR_DESKTOP
    dic = get_wyy_mp3(save_file_name=save_file_name, save_folder_path=save_folder_path, wyy_url=wyy_url)
    if dic.get("k_success") == "False":
        get_wyy_mp3_v2(save_file_name=save_file_name, save_folder_path=save_folder_path, wyy_url=wyy_url)


def main():
    get_mp3()


if __name__ == '__main__':
    print("start test!")
    main()
    print("\n\n\neverything is ok!")



import os
import sys
from retry import retry
import traceback
import pysnooper
import user_agent
import time
import random
import requests
import re
import json
import collections
import threading   # threading.Thread 为可使用的多线程的类！
import multiprocessing # multiprocessing.Process 为可使用的多进程的类！
from scrapy import Selector
from urllib import parse
import smtplib,uuid
from email.mime.text import MIMEText
import execjs
from copy import deepcopy
import redis
import socket

## exchange邮件发送相关的库包
from exchangelib import DELEGATE, Account, Credentials, Configuration, NTLM, Message, Mailbox, HTMLBody
from exchangelib.protocol import BaseProtocol, NoVerifyHTTPAdapter
import urllib3
urllib3.disable_warnings() # 可以避免老是报错...

from kw618.k_requests.abuyun_proxies import AbuyunProxies
from kw618.k_requests.selenium_login import get_one_cookie

from kw618._file_path import *






# 一个traceback的函数, 用于定位bug出现的位置 (防止except Exception as e中, 只报错, 不报位置...有点坑)
k_tb = traceback.format_exc
"""
    使用案例:
        try:
            1/0
        except:
            print(k_tb(10))  ## except中, 可以呈现详细的报错信息
"""



class CookiesPool():
    """
        function:
            封装一个操作redis的对象, 把列表/哈希表/集合等redis数据结构的 '增删改查' 写成统一的'实例方法接口'
            pool的数据结构: 使用队列结构 (先进的先出, 适合爬虫项目)
                            所以: add是加到右边最后一个元素; get是左边第一个元素; pop也是左边第一个元素
        note:
            1. pool_type的种类:
                1. list
                2. dict
                3. set
    """
    def __init__(self, pool_name, pool_type="list", redis_host="localhost", redis_port=6379, db=0, decode_responses=True, **kwargs):
        if db:
            kwargs["db"] = db
        if decode_responses:
            kwargs["decode_responses"] = decode_responses
        self.redis = redis.StrictRedis(host="localhost", port=6379, **kwargs)
        self.pool_name = pool_name
        self.pool_type = pool_type
        # 先清空redis中的这个'键' (如果已经存在某些数据, 可能数据类型会和后面操作的方式对不上而报错)
        self.redis.delete(self.pool_name)

    # 增:
    # =================
    def add(self, value, key="undefined"):
        if self.pool_type == "list":
            res = self.redis.rpush(self.pool_name, value)
        elif self.pool_type == "dict":
            res = self.redis.hset(self.pool_name, key, value)
        return res

    # 删:
    # =================
    def pop(self):
        if self.pool_type == "list":
            res = self.redis.lpop(self.pool_name)
        elif self.pool_type == "dict":
            res = "'dict'类型的pool没有pop方法;\n"
        return res

    def delete(self, key):
        if self.pool_type == "list":
            res = self.redis.lrem(self.pool_name, 1, key)
        elif self.pool_type == "dict":
            res = self.redis.hdel(self.pool_name, key)
        return res

    # 改:
    # =================
    pass


    # 查:
    # =================

    def get(self, key="undefined"):
        if self.pool_type == "list":
            res = self.redis.lindex(self.pool_name, 0)
        elif self.pool_type == "dict":
            res = self.redis.hget(self.pool_name, key)
        return res

    def values(self):
        if self.pool_type == "list":
            res = self.redis.lrange(self.pool_name, 0, -1)
        elif self.pool_type == "dict":
            # "dict类型的, 直接返回items字典"
            res = self.redis.hgetall(self.pool_name)
        return res

    def length(self):
        if self.pool_type == "list":
            res = self.redis.llen(self.pool_name)
        elif self.pool_type == "dict":
            res = len(self.redis.hvals(self.pool_name))
        return res

    def __str__(self):
        return str(self.values())

    def __repr__(self):
        "推荐用这个!"
        return str(self.values())


class myRequest():
    def __init__(self):
        self.name = "kerwin"
        self.session = requests.Session()



    # 所有请求都用这个接口！
    # @log_error()
    @retry(tries=3, delay=4, backoff=2, max_delay=8) # 这里的4次retry是正常的网络连接出错重试(不考虑反爬)
    # @pysnooper.snoop()
    def req(self, url, *, proxies=None, selector=False, auth=None, allow_redirects=True,
            data=None, selenium=False, req_method="get", is_obj=False, use_session=False,
            sleep_sec=None, other_headers={}, timeout=30, encoding=None, requests=requests,
            **kwargs):
        """
        params:
            data: dict类型
            other_headers: dict类型
            **kwargs:
                cookies: 可传入一个dict类型的cookies_dict  (前提: other_headers中不能有"cookie")



        note:
            1.所有请求都用这个通用接口(要做到普遍性)
            2.服务器连接类型的出错,可以直接用上面的retry.不需要在robust_req中考虑了!
            3. 默认情况只返回 res_obj.text 文本 (一般需要用到res_obj对象比较少见)
            4. verify的作用是: 表明在访问https协议的网址时, 是否需要验证服务器的 server's TLS certificate. requests库默认为True.
                有时候使用charles时候在用req访问 'https网址', 会报错: 服务器tls证书验证失败, 导致的SSLError(bad handshake)
            5. requests.get() 方法是默认允许 "重定向/跳转"的!!
                    so: 当需要禁止跳转的时候, 在get()方法中加入: allow_redirects=False

        tips:  res_obj.status_code 为200时再做后续处理？
        """

        headers = {
            "User-Agent": user_agent.generate_user_agent(),
            # "Host": "www.dianping.com",
            # "Referer": "http://price.ziroom.com",
            # "Cookie": "a=123;b=456;c=789",
        }
        # 0. 如果使用session，之后302跳转过程中的所有cookies，都会存储在这个 self.session中
            ### 上面的入参 requests=requests 是避免这里把全局变量requests给屏蔽掉了
        if use_session:
            requests = self.session

        # 1.添加其他必要的头部信息(包括cookies可以以{"Cookie":"a=123;b=456;c=789"}形式update进来)
        # cookies 也可以含在other_headers里面
        headers.update(other_headers)

        # 2.使用代理ip
        if proxies is not None:
            if proxies == "abuyun":
                proxies = AbuyunProxies.proxies
            elif proxies == "clash":
                proxies = {
                    # 代理设置: 如果要翻墙, 需要把端口号改成clash的端口号 (这里的1080是以前ssr的端口)
                    "http" :"http://localhost:7890",
                    "https" :"https://localhost:7890",
                    # "http" :"socks5://127.0.0.1:7891",
                    # "https" :"socks5://127.0.0.1:7891",
                }
        else:
            # 默认使用不使用代理
            proxies = {
                # 代理设置: 如果要翻墙, 需要把端口号改成clash的端口号 (这里的1080是以前ssr的端口)
                # "http" :"http://localhost:1080",
                # "https" :"https://localhost:1080"

                # 需要密码认证的代理:
                # "http" :"http://user:password@ip:port",

                # 也可以用socks协议的代理:
                # "http" :"socks5://user:password@ip:port",
                # "https" :"socks5://user:password@ip:port",

                # notice: 意味着 '左边的请求' 会使用 '右边的代理' 来访问

            }


        # 3. 设置其他'关键参数'的默认值
        kwargs.setdefault('verify', False)



        # 4.开始发送req
        print("\nreq 已发出。。\n")
        if req_method == "get":
            res_obj = requests.get(url, headers=headers,proxies=proxies, auth=auth, timeout=timeout, allow_redirects=allow_redirects, **kwargs)
        # 下面的 data参数 必须要先用json.dumps()转成str后，才能发送请求！！！ //k200628: 不需要转成json啊, 直接用dict类型就能传入请求啊
        elif req_method == "post":
            res_obj = requests.post(url, headers=headers,proxies=proxies, data=data, timeout=timeout, allow_redirects=allow_redirects, **kwargs)
        elif req_method == "put":
            res_obj = requests.put(url, headers=headers,proxies=proxies, data=data, timeout=timeout, allow_redirects=allow_redirects, **kwargs)
        self.res_obj = res_obj

        # 5.修改需要的编码格式
        if encoding:
            res_obj.encoding = encoding

        # 6.当需要返回原始response对象的时候
        if is_obj:
            return res_obj
        else:
            res_text = res_obj.text

        # 7.当需要对页面进行解析的时候可以开启这段代码
        if selector:
            selector = Selector(text=res_text)
            return selector

        # 8.睡眠 (怕被反爬..)
        if sleep_sec:
            time.sleep(sleep_sec)

        # 9.返回文本string
        return res_text


    def get_session_cookies_dict(self):
        # 获取整个session中的所有cookies！！好用！！ (自动解析cookiejar成dict)
        return requests.utils.dict_from_cookiejar(self.session.cookies)

    def get_res_obj_cookies_dict(self):
        return requests.utils.dict_from_cookiejar(self.res_obj.cookies)




    # 该函数为每个网站自定义的部分,需要抓包,多测试,才能让req更贱稳健!
    def robust_req(self, url, proxies=None, selector=False, auth=None,
                    data=None, selenium=False, req_method="get",
                    sleep_sec=None, abnormal_num=0,
                    get_cookies=None, other_headers={}):
        """
        说明:
        1.该函数,可以自由定义.自己决定ip被封的判断条件.这样会使req更加稳健.
        2.该函数必须要起到分类response的功能.(即:必须要划分好准确的my_status)

        重点:
        返回的数据要有两个元素:
        1.目标对象   (具体是str还是json,视情况而定)
        2.my_status
        """

        res_txt = self.req(url, proxies=proxies, selector=selector, auth=auth,
                            data=data, selenium=selenium, req_method=req_method,
                            sleep_sec=sleep_sec, other_headers=other_headers)

        # to be deleted: my_retry 这样的写法好像很鸡肋,不规范!
        # 这个loads时候很容易出错,这边设置重试3次!
        # exec_str = "json.loads(res_txt, strict=False)"
        # res_dict = my_retry(exec_str, 4, error_txt="json数据格式转化")
        res_dict = json.loads(res_txt, strict=False)


        # print(res_dict)

        # if get_cookies:
        #     cookies = get_cookies() # 返回格式为dict

        # 此处为req使用代理ip的重试次数,下面的数字"3"可调!
        for retry_time in range(1, 3):
            # 1. response 为正常情况
            if res_dict.get("status", 0) == "1":
                # res_dict.update({"my_status":"ok"})
                print("\n楼盘数据获取成功\n")
                return res_dict, "ok"

            # 2. response 表示没有找到相关数据,可以把请求放入abandon集合
            elif res_dict.get("status", 0) == "7":
                print("\n没有此楼盘数据,可以abandon..\n")
                return res_dict, "abandon"

            # 如果response都不在上述判断中,则视为ip被封,更改ip代理,重新req一遍
            else:
                # if get_cookies:
                #     other_headers = get_cookies() # 返回格式为dict
                print("\nip被封，更换ip...\n")
                res_txt = self.req(url, proxies="abuyun", selector=selector, auth=auth,
                                    data=data, selenium=selenium, req_method=req_method,
                                    sleep_sec=0.05, other_headers=other_headers) # 使用代理的话不用睡眠很久的!
                # 这个loads时候很容易出错,这边设置重试3次!
                exec_str = "json.loads(res_txt, strict=False)"
                res_dict = my_retry(exec_str, 4, error_txt="json数据格式转化")


        # 如果经过上面多次循环req还是找不到一种""已知归属"",则返回ip错误
        return res_dict, "wrong"


def get_my_ip(option=None, proxies=None):
    "用来测试服务器观测到我的请求是来自哪个ip的"
    "与代理ip相关"
    req = myRequest().req
    api_url = 'http://pv.sohu.com/cityjson?ie=utf-8'
    res_text = req(api_url, proxies=proxies)
    res_text = re.findall(r'({[\s\S]+})', res_text)[0]
    print(res_text)
    res_dict = json.loads(res_text)
    if option == "ip":
        return res_dict.get("cip")
    elif option == "name":
        return res_dict.get("cname")
    else:
        return res_dict






def txt_2_html(txt):
    txt = txt.replace("\n", "<br>")
    txt = txt.replace(" ", "&nbsp;")
    txt = txt.replace("font:", "font ")
    txt = txt.replace("div:", "div ")
    txt = txt.replace("table:", "table ")
    txt = txt.replace("th:", "th ")
    txt_html = txt
    return txt_html







class kEmailSenderExchange():
    "使用outlook邮箱的那种模式发邮件"

    def __init__(self):
        #此句用来消除ssl证书错误，exchange使用自签证书需加上
        BaseProtocol.HTTP_ADAPTER_CLS = NoVerifyHTTPAdapter

        # 配置基本用户信息
        # self.user_name = "lvzc1"
        # self.password = "Lzc15168201914*"

        # self.user_name = "chenb"
        # self.password = "Zz00000000."

        self.user_name = "yangjj4"
        self.password = "Yangjiajing131#"
        self.email_domain = "ziroom.com"
        self.server_address = "zr-casc.ziroom.com"

        # 登录
        self.login()

    def login(self, user_name=None, password=None, email_domain=None, server_address=None):
        """后续应该建立异常捕获机制"""
        user_name = user_name if user_name else self.user_name
        password = password if password else self.password
        email_domain = email_domain if email_domain else self.email_domain
        server_address = server_address if server_address else self.server_address

        # 连接/登录 我的邮箱
        my_email_address = "{}@{}".format(user_name, email_domain)
        cred = Credentials(r'{}\{}'.format(email_domain, user_name), password)
        config = Configuration(server=server_address, credentials=cred, auth_type=NTLM)
        self.account = Account(
            primary_smtp_address=my_email_address, config=config, autodiscover=False, access_type=DELEGATE
        )
        print('我的邮箱已连接/登录...\n')


    def python_str_2_html_tag(self, message, **kwargs):
        # print(kwargs)
        message = message.replace("\n", "<br>")
        message = message.replace(" ", "&nbsp;")
        message = message.replace("font:", "font ")
        message = message.replace("div:", "div ")
        message = message.replace("table:", "table ")
        message = message.replace("th:", "th ")
        # message = message.replace('<tr>', '<tr align="center">')
        # message = message.replace('&lt;', '<')
        # message = message.replace('&gt;', '>')
        message = message
        return message


    def k_send_msg(self, message, subject_="无主题", receiver_lst=['365079025@qq.com'], need_to_conver_html=False, **kwargs):
        # 1. 把message先转换成能在浏览器正常显示的 html类型 的文本
        if need_to_conver_html:
            message = self.python_str_2_html_tag(message)

        # 2. 编辑邮件内容
        email_obj = Message(
            account=self.account,
            subject=subject_,
            body=HTMLBody(message),
            # to_recipients=[Mailbox(email_address=['365079025@qq.com'])]
            to_recipients = [Mailbox(email_address=receiver_address) for receiver_address in receiver_lst]
        )

        # 3. 发送邮件
            ### 随机睡眠: 防止被封
        random_time = random.random()*2
        print("发送邮件前，睡眠 {}  中。。。".format(random_time))
        time.sleep(random_time)
            ### 异常捕获
        try:
            x = email_obj.send()
            print("邮件发送成功！")
            return True
        except Exception as e:
            print(e)
            print("邮件发送失败！")
            return False










class kEmailSenderSmtp():
    "**使用中转模式发邮件"

    def __init__(self):
        # 发邮件相关参数
        self.sender='365079025@qq.com'
        # qq邮箱授权码
        self.password="rrtvphpkphylbgja"
        # 服务提供商的邮箱后缀??
        self.smtpsever='smtp.qq.com'
        # 端口号
        self.port=465
        # 链接服务器发送 (是个发送邮件的实例对象)
        self.smtp = smtplib.SMTP_SSL(self.smtpsever,self.port)
        # 登录
        self.login()

    def login(self):
        #登录
            ### 登录时候容易连接错误报错,应该捕获异常,并重新执行!
        try:
            for count in range(1, 4): # 尝试3次连续登录
                try:
                    print("正在 login 邮箱..")
                    self.smtp.quit() #1. 先退出连接
                    self.smtp.connect(self.smtpsever, self.port) #2. 再连接
                    self.smtp.login(self.sender, self.password)
                    print("邮箱 login 成功..")
                    break # 如果能正常登录, 则退出循环
                except Exception as e:
                    print("尝试第 {} 次login 邮箱..".format(count))
                    time.sleep(1)
                    print(e)
        except Exception as e:
            print(e)
            raise Exception("邮箱连续登录3次, 依旧出错...")


    def python_str_2_html_tag(self, message, **kwargs):
        # print(kwargs)
        message_html = message.replace("\n", "<br>")
        message_html = message_html.replace(" ", "&nbsp;")
        return message_html


    def k_send_msg(self, message, subject_="无主题", receiver_lst=['365079025@qq.com'], **kwargs):
        """
        in: 1. receiver_lst 接受者的邮箱
            2. message为需要发生邮件的正文内容
            3. subject 邮件显示中的第一栏（主题）

        notes:
            1. 这里传送进来的message中的\n换行是python语法中的，而如果要在前端html中展示换行，
                需要使用<br>,或者<div>等html标签

        tips:
            1. 字体大小
                最小： <font size="1">a</font>
                最大： <font size="6">a</font>
            2. 字体颜色
                字体红色： <font color="#ff0000"> a </font>
            3. 背景颜色
                背景颜色黄色：<span style="background-color: rgb(255, 255, 0);"> a </span>

        todo:
            想用**kwargs的关键词传参，来自动使某些需要的元素格式化（字体大小、颜色等）
        """

        # 1. 把msg先转换成能在浏览器正常显示的 html类型 的文本
        message_html = self.python_str_2_html_tag(message)

        # 2. 编辑邮件内容
        body = message_html # 正文内容
        msg=MIMEText(body,'html','utf-8') ## 使用html格式解析器
        msg['from'] = self.sender # 不加这行也可以发送，但是邮箱列表中没有发件人的头像和名称。
        msg['subject'] = subject_

        # 3. 发送邮件
            ### 随机睡眠: 防止被封
        random_time = random.random()*3
        print("发送邮件前，睡眠 {}  中。。。".format(random_time))
        time.sleep(random_time)
            ### 异常捕获
        for try_time in range(3):
            try:
                self.smtp.sendmail(self.sender, receiver_lst, msg.as_string())  #发送
                print("邮件发送成功！")
                return True
            except Exception as e:
                print(e)
                print("邮件发送失败！尝试重新login....")
                self.login()
        return False

    def quit(self):
        #关闭
        self.smtp.quit()























# def k_send_msg(message, subject_="无主题", receiver_lst=['365079025@qq.com'], **kwargs):
#
#     def python_str_2_html_tag(message, **kwargs):
#         # print(kwargs)
#         message_html = message.replace("\n", "<br>")
#         message_html = message_html.replace(" ", "&nbsp;")
#         return message_html
#
#
#     message_html = python_str_2_html_tag(message)
#
#     #发邮件相关参数
#     sender='365079025@qq.com'
#     # password="kwijdwfiqsajcajb"            #qq邮箱授权码
#     # password = "kodcgdnhvcxxbhcc"
#     password="rrtvphpkphylbgja"            #qq邮箱授权码
#     # receiver_lst = ['365079025@qq.com', "962600510@qq.com"]
#     if not receiver_lst:
#         receiver_lst = ['365079025@qq.com']
#     smtpsever='smtp.qq.com'
#     port=465
#
#     #编辑邮件内容
#     body = message_html # 正文内容
#     msg=MIMEText(body,'html','utf-8')
#     msg['from'] = sender # 不加这行也可以发送，但是邮箱列表中没有发件人的头像和名称。
#     msg['subject'] = subject_
#
#     #链接服务器发送
#     smtp = smtplib.SMTP_SSL(smtpsever,port)
#     # 登录时候容易连接错误报错,应该搞个try,捕获异常,并重新执行!
#     smtp.login(sender,password)                          #登录
#     """频繁登录，发送，退出？？？是不是得封装在一个类中会更好一些？？"""
#     """而且发送邮件的频率要限制一下，给一个随机睡眠吧。。。"""
#     random_time = random.random()*3
#     print("发送邮件前，睡眠 {}  中。。。".format(random_time))
#     time.sleep(random_time)
#     smtp.sendmail(sender, receiver_lst, msg.as_string())  #发送
#     smtp.quit()                                     #关闭
#     print("邮件发送完成！")
#
#
#     """
#     发送邮件的异常捕获机制要搞起来！！啥时候没发送成功自己要知道
#     """
#
#

# msg = 'name:aa\nage ： 99\n\n   hhh'
# k_send_msg(msg, receiver_lst=["lvzc1@ziroom.com"])
# sys.exit()







def _robust_req(url):
    """
    note:
        1.该下划线开头的robust_req,仅作为模板使用.
          以后可以在此基础上做局部修改,套用在其他项目上.
        2.此函数是在req的基础上做了定制化处理,
          考虑到多种可能的返回值,分流处理,稳定地得到最终需要的数据.
    """
    # 0. 先获取redis的连接对象
    r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)

    for retry_time in range(1, 4):
        try:
            # 1.正常发起req
            token = r.get("price_system:AccessToken")
            other_headers = {
                "Access-Token":token,
            }
            res_txt = req(url, other_headers=other_headers)
            res_dict = json.loads(res_txt, strict=False) # 此处,我知道正常情况他会返回json格式,才这么写(需要针对不同场景写)

            # 2.如果被反爬, cookies验证失败，重新获取cookies
            if res_dict.get("code", 0) == "40002":
                print("token验证失败，使用post模拟登陆，获取cookies....")
                token = PriceSysCookies().get_target_cookie()
                r.set("price_system:AccessToken", token)
                print("\n最新token:{}\n".format(token))
                other_headers = {
                    "Access-Token":token,
                }
                res_txt = req(url, other_headers=other_headers)
                res_dict = json.loads(res_txt, strict=False)
            # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> (以上是关于cookies的验证)

            # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> (以下是关于response的分流处理)
            # 1..如果获取到的status字段=7,说明服务器没有找到我想要的数据,遗弃它
            if res_dict.get("status", 0) == "7":
                return {}, "abandon"
            # 2.如果获取到的status字段=5,说明发生其他问题,待处理
            elif res_dict.get("status", 0) == "5": #此处占坑,作为其他错误response的捕获#
                return {}, "failure"
            # 3.如果以上错误response都没捕获到,说明response正常,返回该正常的response
            else:
                return res_dict, "success" #"被期望"的正常情况#

        except Exception as e:
            print(e)
            ## 服务器确实只返回这样的数据,可能是被反爬了,也可能是发送给服务器的数据有问题
            ## (大概率是被反爬了:1.ip被封;2.cookies过期;3.headers/query缺失等)
            ## 需要在上面代码块重新调整"返回值的判断逻辑"!!
            print("response 非目标对象, 可能被反爬!")
            print("马上进行第 {} 次req重试\n[睡眠3秒]".format(retry_time))
            time.sleep(3)
            continue
        return {}, "error"


# 生产者
def k_produce(lst=[], queue_name="unnamed", sub_queue_name="", produce_type="init"):
    """
    获取需要爬取的所有id,清空并存入redis的待爬队列中
    """
    # 0. 先获取redis的连接对象
    r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)

    # 1. 先把传入的队列名称加上 ":" (方便后面redis中分级)
    if queue_name:
        queue_name += ":"
    if sub_queue_name:
        sub_queue_name += ":"


    # 2. 分流: 区分生产者类型
        ### 1. 清空原有e, 重新放入所有e
    if produce_type == "init":
        # 将lst中的所有 '元素' 放入待爬队列
        # 1. 先清空历史遗留的元素
        r.delete("{0}{1}pending_queue".format(queue_name, sub_queue_name))
        # 2. push所有元素
        if len(lst): # 如果lst没有元素,下面的lpush会报错
            r.lpush("{0}{1}pending_queue".format(queue_name, sub_queue_name), *lst)
        ### 2. 加入新的e
    elif produce_type == "add":
        if len(lst): # 如果lst没有元素,下面的lpush会报错
            r.lpush("{0}{1}pending_queue".format(queue_name, sub_queue_name), *lst)




def k_consume_pool(consume_func, queue_name="unnamed", sub_queue_name="", thread_num=5, time_sleep=0.3):
    """
    function: 使用多线程，消费redis队列中的所有内容
                (消费方式需用用consume_func函数来定义)
                (可以是发起req，也可以是其他操作...)

    既然是多线程,一般需要用for循环,循环生成线程.
    再把这些子线程统一append到一个t_pool中管理.
    join()用于主线程阻塞等待这些子线程
    """
    r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)

    #################################### 分割线 ##########################################
    def safety():
        """
        function: 确保不发生漏爬现象
        非常完整的req流程，可供多线程并发爬取，不会存在抢占资源的问题
        """

        # 如果待爬队列有元素,就一直循环爬取
        while r.llen("{0}{1}pending_queue".format(queue_name, sub_queue_name)):
            num = r.llen("{0}{1}pending_queue".format(queue_name, sub_queue_name))
            print("待爬队列中的剩余元素量:{}\n".format(num))
            # 1.从待爬队列获取出e
            e = r.rpop("{0}{1}pending_queue".format(queue_name, sub_queue_name))
            # 2.将 e 放入error_queue
            r.lpush("{0}{1}error_queue".format(queue_name, sub_queue_name), e)

            try:
                # 3.真正执行任务的函数！！！（消费者函数）
                consume_func(e)
                # 4.将成功爬取的url存入已爬队列
                r.sadd("{0}{1}crawled_queue".format(queue_name, sub_queue_name), e)
                # 5.删掉错误队列中的响应元素
                r.lrem("{0}{1}error_queue".format(queue_name, sub_queue_name), -1, e)

            # 发送req后,不管出现什么问题,都把url遗留在error_queue中
            except Exception as e:
                print(e)
                tb_txt = traceback.format_exc(limit=5)
                print(tb_txt)
                # 如果req访问出错,数据获取失败,不管什么原因,把这个错误id遗留在错误队列中
                print("e={} 访问出错,遗留在错误队列中\n\n".format(e))
    #################################### 分割线 ##########################################

    # 1. 先把传入的队列名称加上 ":" (方便后面redis中分级)
    if queue_name:
        queue_name += ":"
    if sub_queue_name:
        sub_queue_name += ":"

    # 2.
    t_pool = []
    for _ in range(thread_num): # 此处创建5个线程来消费
        print("\n即将生成一个线程...")
        time.sleep(time_sleep)
        t = threading.Thread(target=safety)
        t.start()
        t_pool.append(t)
    for t in t_pool:
        t.join() # 主线程需要阻塞等待每一个子线程



def k_error_to_pending(queue_name):
    """
    function : 把error队列的数据转移到待爬队列中,便于开始对error的req重新爬取
    """
    # 0. 获取redis的连接对象
    r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)

    # 1. 翻转error队列
    error_lst = r.lrange("{0}:error_queue".format(queue_name), 0, -1)
    error_lst.reverse()

    # 2. 把error队列中的错误元素, 倾倒入待爬队列中
    if len(error_lst): # 如果lst没有元素,下面的lpush会报错
        r.lpush("{0}:pending_queue".format(queue_name), *error_lst)
    # 3. 删除error队列
    r.delete("{0}:error_queue".format(queue_name))


def kill_error_queue(queue_name):
    "清空 error_queue 的元素"
    r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)
    r.delete("{queue_name}:error_queue".format(queue_name=queue_name))


def k_crawl_error_again(queue_name, consume_func, thread_num=1, time_sleep=1, try_loops=1):
    "如果有错误req导致error_queue有留存元素,将其倒入pending_queue队列,重新对该部分爬取"
    r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)
    for try_times in range(try_loops): # 重试2次
        # 如果error_queue队列中已经没有元素,则跳出循环
        error_lst = r.lrange("adjust_price:error_queue", 0, -1)
        if len(error_lst) == 0:
            break
        # 否则倾倒队列,重新生成多线程消费者消费
        k_error_to_pending(queue_name=queue_name)
        k_consume_pool(queue_name=queue_name, consume_func=consume_func, thread_num=thread_num, time_sleep=time_sleep)



def get_js_txt_for_encrypt(js_full_path):
    """
    in: js代码的完整路径
        （经过自己过滤处理后，把所有可能用到的加密js函数全放在这个js文件中即可！！）
        必须要有个直接调用加密的函数
            eg.
            function encryptByDES(message, key) {.....}

    note:
        其实只要找到所有需要的js文件，直接复制、粘贴过来，都不需要修改什么，直接就可以用了诶！！！
        目前的难点是：1. 如何找到加密函数？   2. 如何把与加密相关的所有js文件都集齐、汇总到本js文件中
        难点1的思路：
            1. 使用事件断点：  click、xhr、encode解析、submit 等等
            2. 直接搜索关键词： encrypt, aes, submit, request, click, password, user, passpd, log, login
            3. 针对登录的form、输入密码的input等标签，后续js加密代码肯定需要使用这个标签获取这个input中的值，所以可以搜这个tag的id或类名
            4. 设置很多可能的断点，逐行逐行调试。配合使用堆栈的记录功能，快速跳转函数。(时刻关注local的变量值的变化)
        难点2的思路：
            1. 多配合"step into next function call" 的调试方式（可以下钻到更深一层的函数）
    """
    with open(js_full_path, "r") as file:
        js_txt = file.read()
    return js_txt




def exec_js_function(js_full_path, func_name, *args):
    """
    function: 获取js代码的上下文，指定某个js函数，并传入相应个数的参数，得到js函数的返回值
    参数说明:
        1.解密js的完整路径
        2.所使用的函数
        3+.可以传入多个参数
    """
    # print(args)
    # print("---\n\n")
    # print(*args)
    js_txt = get_js_txt_for_encrypt(js_full_path)
    ctx = execjs.compile(js_txt, f"{FILE_PATH_FOR_HOME}/node_modules")
    result = ctx.call(func_name, *args)
    return result


# 测试一 : 价格系统
# js_full_path = "/Users/kerwin/MyBox/MyCode/Personal/django_test01/first_project/static/js/encryption_4_price_system.js"
# a = exec_js_function(js_full_path, "encryptByDES", "Lzc15168201914*", "#z@i!r*o%o&m^")
# print(a)
# 测试二 : 内网（库存系统）
# js_full_path = "/Users/kerwin/MyBox/MyCode/Personal/django_test01/first_project/static/js/encryption_4_ziroom_intranet.js"
# a = exec_js_function(js_full_path, "encodeAes", "Lzc15168201914*")
# print(a)
# sys.exit()




# def k_urlEncode(txt):
def url_encode(txt):
    "url编码: 网络请求中特有'编码格式'(区别于unicode);"
    "正常来说, http/s请求的时候, 会自动把中文进行url编码; 但有时候, 需要我们手动对url进行编码, 所以写了个k接口"
    return parse.quote(txt)

# def k_urlDecode(txt):
def url_decode(txt):
    "url解码: 把%E6%9c%88 解码成 '月' "
    return parse.unquote(txt)






def main():
    # url = "https://www.amap.com/service/poiTipslite?&city=330100&type=dir&words={}".format("芳满庭")
    # res_txt = myRequest().robust_req(url)
    # print(res_txt)
    # print(dir())
    # print(__file__)
    # print(locals())
    print("ssssssssss"*88)





if __name__ == '__main__':
    print("start test!")
    main()
    print("\n\n\neverything is ok!")



import json
import pandas as pd


# 导入常用的固定路径(多平台通用)
from kw618._file_path import *
# 本脚本依赖很多 utils_requests的函数和模块, 直接用*  (注意要避免循环导入问题)
from kw618.k_requests.utils_requests import *
from kw618.k_requests.ocr import *
from kw618.k_python.utils_python import *
from kw618.k_pandas.utils_pandas import *

from kw618.k_finance.utils_quant import *
from kw618.k_finance.const import ulist_change_name_dict

req = myRequest().req
client = pymongo.MongoClient(f'mongodb://kerwin:kw618@{HOST}:27017/')
db_for_quant = client["quant"]




# 原始模型类
class OrgModel():
    # """
    #     hc模型中, 暂时只允许每次只hc一个stock;
    #     跑sp的时候, 只要给每个stock一定的初始cash_balance即可
    # """

    def __init__(self, stock_code="", Zqdm="", init_cash_balance=100000):
        """
            params:
                init_cash_balance: 初始账户余额

            notice:
                1. 初始化时, Zqdm和stock_code 二选其一, 就可以实现初始化.

        """
        # 1. 初始化'静态'数据
        self.init_cash_balance = init_cash_balance # 记录一个初始启动资金(固定)
        self.cash_balance = self.init_cash_balance # 后面动态更新的余额
        if stock_code:
            self.stock_code = stock_code
            self.Zqdm = Q._StockCode_to_Zqdm(stock_code_lst=[self.stock_code])[0]
        elif Zqdm:
            self.Zqdm = Zqdm
            self.stock_code = Q._Zqdm_to_StockCode(Zqdm_lst=[self.Zqdm])[0]
        else:
            raise Exception("代码输入错误, 初始化未成功!!\n")
        self.Zqmc = Q._Zqdm_to_Zqmc(Zqdm_lst=[self.Zqdm])[0]
        self.f1 = Q._Zqdm_to_f1(Zqdm_lst=[self.Zqdm])[0] # 2:A股; 3:ETF基金;  (后期可用于计算交易费的佣金差异)
        self.sxf_rate = 0.00018 # 手续费率按'万1.8'来计算
        self.ETF_fee_rate = 0.006/365 # 托管费率按 '一年千6' 来计算 (托管费每日计提, 无需在每笔交易中另行支付) (貌似只有在及哦啊一日才会生成)

        # 2. 读取mongo中的历史'日K'数据的df
            # (每个实例都需要获取新的'日K'数据) (后面跑策略依赖这里的'日K'数据!!)
        whole_hist_df = read_mongo(db_for_quant["hist_allstock"], {"stock_code":self.stock_code})

        # 3. 数据预处理 (得到双均线的差值)
        # whole_hist_df["_num"] = np.arange(1, len(whole_hist_df)+1) # 新建一个'顺序序列' (但是需要在init_data函数中操作此步骤)
        whole_hist_df["5v30"] = whole_hist_df["5d_avg"] - whole_hist_df["30d_avg"]
        whole_hist_df["5v20"] = whole_hist_df["5d_avg"] - whole_hist_df["20d_avg"]
        whole_hist_df["5v10"] = whole_hist_df["5d_avg"] - whole_hist_df["10d_avg"]
        whole_hist_df["10v30"] = whole_hist_df["10d_avg"] - whole_hist_df["30d_avg"]
        self.whole_hist_df = whole_hist_df # 所有完整的历史数据
        self.launch_date = self.whole_hist_df.iloc[0]["market_date"] # 上市日期

        # 4. 初始化'动态'数据
        self.init_data()


    def init_data(self, selected_days=99999, start_date=None, end_date=None, pt=False):
        """
            function: 用于后面重复执行run方法时候使用的'初始化'方法 (区别于__init__)
                    步骤:
                        1. 筛选回测日期
                        2. 初始化'通用型变量'
                        3. 执行'do_every_evening'的汇总函数
                        4. 初始化'针对型变量'
            params:
                selected_days: 筛选的天数  (99999代表: 取所有日期)
        """

        # 1. 筛选回测日期 (得到query_set)
            # i. 若给了一个start_date和一个selected_days: 从开始日期向后取xx天
            # ii. 若给了一个end_date和一个selected_days: 从截止日期向前取xx天
        if start_date is not None:
            if end_date:
                pass
            else:
                end_date = get_later_date(start_date, f"{selected_days} days")
        elif start_date is None:
            if end_date:
                start_date = get_previous_date(end_date, f"{selected_days} days")
            else:
                end_date = get_today_date()
                start_date = get_previous_date(end_date, f"{selected_days} days")
            # 如果开始日期早于'上市日期', 则把'上市日期'作为'开始日期'
        if start_date < self.launch_date:
            start_date = self.launch_date # 覆盖它
            # 如果结束日期晚于'今天最新日期', 则把'今天日期'作为'结束日期'
        if end_date > get_today_date():
            end_date = get_today_date() # 覆盖它
        query_set = f"'{start_date}' <= market_date <= '{end_date}'"
        self.hist_df = self.whole_hist_df.query(query_set)
        self.hist_df["_num"] = np.arange(1, len(self.hist_df)+1) # 新建一个'顺序序列'
        self.selected_days = get_delta_days(start_date=start_date, end_date=end_date)
        self.start_date = start_date
        self.end_date = end_date

        # 2. 初始化'通用型变量':
            # 1. 现金流&开平仓数流
                # (有方向): 历次buying&selling的价格/股数
                # 负数:表示买入该stock; 正数:表示卖出该stock (因为最终统计的是现金的剩余量)
        self.cash_flow_lst = []
        self.deal_count_flag = []
            # 4. 持仓股票数
        self.chicang_stock_count = 0
            # 5. 平均成本价
        self.avg_cost_price = 0
        self.cost_market_value = 0
        self.market_value_percentage = 0
        self.market_value_percentage_lst = []
            # 6. 初始化余额 [临时]
        self.cash_balance = self.init_cash_balance
            # 7. 交易次数
        self.buy_times = 0
        self.sell_times = 0
        self.buy_day_lst = []
        self.sell_day_lst = []
                # 2. 每次交易操作时, 距离上次操作的天数 (不管是买,还是卖)
        self.deal_day_lst = []
            # 8. 是否打印(必要的节点信息)
        self.pt = pt
            # 9. 盈亏比例
        self.LjYkbl = np.nan
        self.DrYkbl = np.nan
            # 11. 更新价格
                # 截取到的时间段内最新一天的价格
        if len(self.hist_df) > 0:
            self.newest_date = self.hist_df.iloc[-1]["market_date"]
            self.newest_price = self.hist_df.iloc[-1]["end_price"]
            # self.newest_price = Q.req_newest_data([self.stock_code])["NewestPrice"].iloc[0]
                    # 暂时用'newest_price'替代 (后续策略计算时, 需要用当天价格覆盖它)
            self.today_date = self.newest_date
            self.today_price = self.newest_price

        #     # 12. 持仓股若卖出需要多少手续费?  # 已经放在 do_every_evening中了
        # self.chicang_stock_deal_fee = 0
        self.ETF_today_fee = 0
        self.ETF_fee_lst = []
        self.ETF_total_fee = 0

        # 3. 执行'do_every_evening'的汇总函数:
        self.do_every_evening()

        # 4. 初始化'针对型变量':
            # 具体策略中需要用到的初始化变量
            # 1. 双均线策略:
                # 上一天的'长短线差值'
        self.last_diff_score = np.nan

            # 2. 止损策略:
                # 累计涨幅:
                    # 用于计算每天的累计涨幅 (是市场的实盘涨幅, 与上面LjYkbl的个人涨幅无关)
                    # (每天append进'最新的累计涨幅')
                # 1. "市场"累计盈亏
        self.MarketLjYk = 0.0  # 这是指"市场上"的"累计盈亏" (区别于: "自己账户"的"LjYk")
        self.MarketLjYkbl = 0.0  # 这是指"市场上"的"累计盈亏比例" (区别于: "自己账户"的"LjYkbl")
        self.MarketLjYk_lst = []
        self.MarketLjYkbl_lst = []
                # 2. "个人"累计盈亏
        self.MyLjYk = 0.0  # 这是指"个人账户"的"累计盈亏" (区别于: "市场"的"LjYk")
        self.MyLjYkbl = 0.0  # 这是指"个人账户"的"累计盈亏比例" (区别于: "市场"的"LjYkbl")
        self.MyLjYk_lst = []
        self.MyLjYkbl_lst = []







    def do_every_morning(self, doc):
        """
            function: 每天早上记录必要的数据
                    (因为不管哪个策略都需要有这一步, 所以把它抽离出来)

            params:
                doc: apply函数中的doc  (也就是self.hist_df中的每一行)


        """

        # 1. 记录 "当天" 的 "日K" 数据
        self.today_date = doc.get("market_date") # 每天实时更新'最新价格'
        self.today_price = doc.get("end_price")
        self.today_days = doc.get("_num") # 得到今天距离'基准日期'的天数
        self.DrYkbl = round(doc.get("growth_rate"), 4) # 每天实时更新'当日盈亏比例'
        self.DrYk = round(doc.get("growth_amount"), 4) # 每天实时更新'当日盈亏'
        if np.isnan(self.DrYkbl) == True: # 如果DrYkbl是空值, 默认用0来填充
            self.DrYkbl = 0.0
        if np.isnan(self.DrYk) == True: # 如果DrYkbl是空值, 默认用0来填充
            self.DrYk = 0.0
        print(self.today_date, self.today_days, self.today_price, self.DrYkbl, self.DrYk)

        # 2. 记录"市场上"的累计数据
        self.MarketLjYk += self.DrYk # 每天累加
        self.MarketLjYkbl += self.DrYkbl # 每天累加
        self.MarketLjYk = round(self.MarketLjYk, 4)
        self.MarketLjYkbl = round(self.MarketLjYkbl, 4)
        self.MarketLjYk_lst.append(self.MarketLjYk)
        self.MarketLjYkbl_lst.append(self.MarketLjYkbl)
        print(self.MarketLjYk, self.MarketLjYkbl)

        # 3. 记录"个人账户"的累计数据
        if self.avg_cost_price != 0:
            self.MyLjYk = round(self.today_price - self.avg_cost_price, 4) # 个人账户的'累计盈亏'的计算方式: 直接用 "最新价格 - 成本均价"
            self.MyLjYkbl = round(self.MyLjYk / self.avg_cost_price, 4)
        elif self.avg_cost_price == 0:
            self.MyLjYk = 0.0
            self.MyLjYkbl = 0.0
        self.MyLjYk_lst.append(self.MyLjYk)
        self.MyLjYkbl_lst.append(self.MyLjYkbl)
        print(self.MyLjYk, self.MyLjYkbl)

        print("\n\n")


    def do_every_evening(self):
        """
            function: 每天晚上记录必要的数据 (一般是受今日交易数据所影响的数据)
                        (相当于对今天数据的汇总)

            params:
                无: 不依赖于doc, 所以也可以放在 init_data()方法 中执行


        """

        # 1. 计算当前持仓股票的'成本市值'
            # 成本市值: 指持有的这些股票共'花了多少钱','最新市值'要大于它才算有利润!
        self.cost_market_value = self.chicang_stock_count * self.avg_cost_price
        # 2. 计算当前持仓股票的'今日市值'
            # i. A股直接相乘就是'今日市值'
        self.today_market_value = self.chicang_stock_count * self.today_price
            # ii. ETF需要在'今日市值'中计提每日的管托费
        if self.f1 == 3:
            self.ETF_today_fee = self.today_market_value * self.ETF_fee_rate
            self.ETF_fee_lst.append(self.ETF_today_fee)
            self.ETF_today_fee = sum(self.ETF_fee_lst)
            self.today_market_value -= self.ETF_total_fee

        # 3. 现金流汇总
        self.sum_cash_flow = sum(self.cash_flow_lst)
        # 4. 开平仓数流汇总
        self.sum_deal_count_flag = sum(self.deal_count_flag)
            # 持有A股, 如果要卖出的话, 需要多少手续费? (ETF直接返回0)
        self.chicang_stock_deal_fee = self.get_commission_fee(
            Mmlb="S", deal_price=self.today_price, deal_count=abs(self.sum_deal_count_flag),
        )

        # 5. 今天的"总资产"
        self.total_capital_value = self.cash_balance + self.today_market_value
        # 6. 今天的"净资产" (已扣除今天持仓的交易成本) (即:假设今天把持有的股票全卖了, 账户总资产是多少?)
        self.net_capital_value = self.total_capital_value - self.chicang_stock_deal_fee
        # 7. 今天的"市值占比"
        self.market_value_percentage = self.today_market_value / self.total_capital_value
        self.market_value_percentage_lst.append(self.market_value_percentage)



    def get_commission_fee(self, Mmlb, deal_price, deal_count):
        """
            params:
                Mmlb: 买卖类别  # "B" / "S"
                deal_price: 交易单价 (一股的价格)
                deal_count: 交易数量
        """

        if deal_count > 0:
            # 会产生手续费
                # 1. A股的手续费
            if self.f1 == 2:
                sxf = max(self.sxf_rate*(deal_price*deal_count), 5.0)

                if Mmlb == "B":
                    total_fee = sxf
                elif Mmlb == "S":
                    yhs = 0.001*(deal_price*deal_count) # 印花税都是'千分之一'的
                    sxf = 5.0
                    total_fee = yhs + sxf

                return total_fee

                # 2. ETF的手续费
            elif self.f1 == 3:
                return 0 # ETF的托管费是'每日计提'的, 所以不再在每笔交易中另外产生
        else:
            return 0


    def buying(self, buying_price, buying_count):
        # 1. 得到真实的所有'买入成本' (包含所有手续费)
        buying_count = buying_count // 100 * 100
        commission_fee = self.get_commission_fee(
            Mmlb="B", deal_price=buying_price, deal_count=buying_count,
        )
        all_buying_cost = buying_price*buying_count + commission_fee # 真实股票市值成本 = 股票成本市值 + 所有的交易手续费

        # 2. [暂时策略]: 账户余额<待购入股票金额, 不断尝试缩小购买量 (尝试3次)
        if self.cash_balance < all_buying_cost:
            for _ in range(3):
                buying_count = int(buying_count *0.98) # 这个0.98可以自定义调整
                all_buying_cost = buying_price*buying_count + commission_fee # 真实股票市值成本 = 股票成本市值 + 所有的交易手续费
                if self.cash_balance >= all_buying_cost:
                    break # 当不断缩小购买量后, 一旦余额足够支付, 则退出循环


        # 3. 只有当现金余额>=待购入股票金额, 才能真的执行'买的操作' (否则就是逻辑bug)
        if (self.cash_balance >= all_buying_cost) and (buying_count != 0):
            # 1. '现金余额'减少
            self.cash_balance -= all_buying_cost
            # 2. '持仓股数'增加
                # (old_cost_market_value 需要在'持仓数量'增加之前计算)
            old_cost_market_value = self.chicang_stock_count * self.avg_cost_price
            self.chicang_stock_count += buying_count
            # 3. '现金流'&'开平仓数流'记录
                # 1. 负现金流: 表示买入'xxx元'的stock
            self.cash_flow_lst.append(-1 * all_buying_cost)
                # 2. 负开平仓数流: 表示买入'xxx股'的stock
            self.deal_count_flag.append(-1 * buying_count)
            # 4. 每次"买入交易"都需要重新计算'平均成本价'  ('卖出交易'不影响平均成本, 只要减去'持仓数量', 计算出来的'成本市值'就会跟着减少了)
                # (成本市值的统计是'无意义'的, 都是动态的用 当前'平均成本价*持仓数量' 计算得到的)
            new_cost_market_value = all_buying_cost
            self.avg_cost_price = (old_cost_market_value + new_cost_market_value) / self.chicang_stock_count
            # 5. '买入次数' +1
            self.buy_times += 1
            # 6. '距离上次买入天数' 添加进 buy_day_lst 中
            self.buy_day_lst.append(self.today_days-sum(self.buy_day_lst))
            # 7. '距离上次交易天数' 添加进 deal_day_lst 中
            self.deal_day_lst.append(self.today_days-sum(self.deal_day_lst))
            if self.pt == True:
                print(f"[{self.today_date}] 买入股票:{buying_count}股, 成本市值:{all_buying_cost}\n", "=="*50)


    def selling(self, selling_price, selling_count):

        # 1. 得到真实的卖出后的'到手现金' (已经扣除所有手续费)
        selling_count = selling_count // 100 * 100
        commission_fee = self.get_commission_fee(
            Mmlb="S", deal_price=selling_price, deal_count=selling_count,
        )
        all_selling_cash = selling_price * selling_count - commission_fee # 到手现金 = 股票卖出价 - 所有的交易手续费

        # 2. 只有当自己的持仓股数大于你要卖的股数时, 才能真的执行'卖的操作' (否则就是逻辑bug)
        if (self.chicang_stock_count >= selling_count) and (selling_count != 0):
            # 1. '现金余额'增加
            self.cash_balance += all_selling_cash
            # 2. '持仓股数'减少
            self.chicang_stock_count -= selling_count
            # 3. '现金流'&'开平仓数流'记录
                # 1. 正现金流: 表示卖出'xxx元'的stock
            self.cash_flow_lst.append(1 * all_selling_cash)
                # 2. 正开平仓数流: 表示卖出'xxx股'的stock
            self.deal_count_flag.append(1 * selling_count)
            # 4. '卖出交易'不会影响到成本价
            pass
            # 5. '卖出次数' +1
            self.sell_times += 1
            # 6. '距离上次卖出天数' 添加进 sell_day_lst 中
            self.sell_day_lst.append(self.today_days-sum(self.sell_day_lst))
            # 7. '距离上次交易天数' 添加进 deal_day_lst 中
            self.deal_day_lst.append(self.today_days-sum(self.deal_day_lst))
            if self.pt == True:
                print(f"[{self.today_date}] 卖出股票:{selling_count}股, 到手现金:{all_selling_cash}\n", "=="*50)


    def return_asset_dict(self):

        # 1. 打印账户整体的资金状况
        if self.pt == True:
            sprint(cash_balance=self.cash_balance)
            sprint(today_market_value=self.today_market_value)
            sprint(net_capital_value=self.net_capital_value)
            print("=="*50)
            print("=="*50, "\n\n\n\n\n")

        # 2. '账户资产'相关信息
        asset_dict = {
            "net_capital_value" : self.net_capital_value,
            # "all_arg_msg" : self.all_arg_msg,
            "stock_code" : self.stock_code,
            "Zqdm" : self.Zqdm,
            # "Zqmc" : self.Zqmc,
            "selected_days" : self.selected_days,
            "launch_date" : self.launch_date,
            "start_date" : self.start_date,
            "end_date" : self.end_date,
            # "sell_day_lst" : self.sell_day_lst,
            "sell_times" : self.sell_times,
            # "cash_flow_lst" : self.cash_flow_lst,
            "sum_cash_flow" : self.sum_cash_flow,
            "cash_balance" : self.cash_balance,
            "today_market_value" : self.today_market_value,
        }

        return asset_dict


    def strategy_cal(self, doc):
        """
            function: 策略计算的主函数 (最主要的"计算模块") [apply函数]
        """
        # 1. 每早:获取必要数据
        self.do_every_morning(doc) # 每个策略都需要更新的'每日数据'

        # 2. 策略的主要内容
        # ====================================================
        # ====================================================
        #     # 1. 金叉: (昨天<0, 今天>0): 需要 buying it
        # if today_diff_score>0 and self.last_diff_score<=0:
        #     buying_count = self.cash_balance // self.today_price # [暂时设定] (待优化)
        #     self.buying(buying_price=self.today_price, buying_count=buying_count)
        #
        #     # 2. 死叉: (昨天>0, 今天<0): 需要 selling it
        # if today_diff_score<0 and self.last_diff_score>=0:
        #     selling_count = self.chicang_stock_count # 当出现死叉时, 把所有的持仓全部抛出
        #     self.selling(selling_price=self.today_price, selling_count=selling_count)

        pass # 每天都不进行任何操作

        # ====================================================
        # ====================================================


        # 3. 每晚:更新必要数据
        self.do_every_evening() # 每个策略都需要更新的'每日数据'


    def run(self, selected_days=99999, start_date=None, end_date=None, pt=False):
        """
            function: 整个'策略类'的执行入口
            params:
                selected_avg: 选择哪两根均线作为长短线的计算差值
                selected_days:
                    1. 回测的天数 (越久远的数据可能与当前的市场差异很大, 而且也伴随通货膨胀之类的影响因素)
                    2. 日期的默认选取方式: 从今天开始, 倒退几天
                    3. 这里的天数: 是指实际天数 (不是"股市的交易天数")
                start_date/end_date: 日期的str标准格式: 2020-07-20
                pt: 是否需要打印输出

            return:
                self.asset_dict (整体的收益情况) (dict类型)

        """

        # 1. 初始化数据(清零)
            # (不清零的话, 会导致重复run方法时候出错)
        self.init_data(selected_days=selected_days, start_date=start_date, end_date=end_date, pt=pt)

        # A. 当存在'该时间段内的df'时:
        if len(self.hist_df) > 0:
            self.all_arg_msg = f"【参数】==> stock_code:{self.stock_code}; days:{self.selected_days} (start:{self.start_date} -- end:{self.end_date})"
            if self.pt == True:
                print("\n\n", self.all_arg_msg, "\n")

            # 2. 执行策略计算
            self.hist_df.apply(self.strategy_cal, axis=1) # 运用上面的"策略计算"函数 (会把所有"买卖操作"数据存入 self.cash_flow_lst)

            # 3. 查看收益情况
            self.asset_dict = self.return_asset_dict() # 返回"账户资产相关信息"

            return self.asset_dict

        # B. 当不存在'该时间段内的df'时:
        else:
            return {} # 返回空dict






# 双均线策略: 回测版本
class DoubleMovingAverage(OrgModel):

    def strategy_cal(self, doc):
        """
            function: 策略计算的主函数 (最主要的"计算模块") [apply函数]
        """
        # 1. 每早:获取必要数据
        self.do_every_morning(doc) # 每个策略都需要更新的'每日数据'
        today_diff_score = doc.get(self.selected_avg) # 可以选取不同的 "长短线" (由run方法中给出)

        # 2. 策略的主要内容
        # ====================================================
        # ====================================================
            # 1. 金叉: (昨天<0, 今天>0): 需要 buying it
        if today_diff_score>0 and self.last_diff_score<=0:
            buying_count = self.cash_balance // self.today_price # [暂时设定] (待优化)
            self.buying(buying_price=self.today_price, buying_count=buying_count)

            # 2. 死叉: (昨天>0, 今天<0): 需要 selling it
        if today_diff_score<0 and self.last_diff_score>=0:
            selling_count = self.chicang_stock_count # 当出现死叉时, 把所有的持仓全部抛出
            self.selling(selling_price=self.today_price, selling_count=selling_count)

        # ====================================================
        # ====================================================


        # 3. 每晚:更新必要数据
        self.do_every_evening() # 每个策略都需要更新的'每日数据'
            # 用今日的score替换昨天的 (保证每个循环中, last_diff_score中都是昨日的数据)
        self.last_diff_score = today_diff_score






    def run(self, selected_avg="5v30", selected_days=99999, start_date=None, end_date=None, pt=False):
        """
            function: 整个'策略类'的执行入口
            params:
                selected_avg: 选择哪两根均线作为长短线的计算差值
                selected_days:
                    1. 回测的天数 (越久远的数据可能与当前的市场差异很大, 而且也伴随通货膨胀之类的影响因素)
                    2. 日期的默认选取方式: 从今天开始, 倒退几天
                    3. 这里的天数: 是指实际天数 (不是"股市的交易天数")
                start_date/end_date: 日期的str标准格式: 2020-07-20
                pt: 是否需要打印输出

            return:
                self.asset_dict (整体的收益情况) (dict类型)

        """

        # 1. 初始化数据(清零)
            # (不清零的话, 会导致重复run方法时候出错)
        self.init_data(selected_days=selected_days, start_date=start_date, end_date=end_date, pt=pt)

        # A. 当存在'该时间段内的df'时:
        if len(self.hist_df) > 0:
            self.selected_avg = selected_avg # 长短线指标选择 (几日均线与几日均线的差值)
            self.all_arg_msg = f"【参数】:stock_code:{self.stock_code}; days:{self.selected_days} (start:{self.start_date} -- end:{self.end_date})"
            if self.pt == True:
                print("\n\n", self.all_arg_msg, "\n")

            # 2. 执行策略计算
            self.hist_df.apply(self.strategy_cal, axis=1) # 运用上面的"策略计算"函数 (会把所有"买卖操作"数据存入 self.cash_flow_lst)

            # 3. 查看收益情况
            self.asset_dict = self.return_asset_dict() # 返回"账户资产相关信息"

            return self.asset_dict

        # B. 当不存在'该时间段内的df'时:
        else:
            return {} # 返回空dict







class StopLoss(OrgModel):

    def strategy_cal(self, doc):
        """
            function: 策略计算的主函数 (最主要的"计算模块") [apply函数]
        """
        # 1. 每早:获取必要数据
        self.do_every_morning(doc) # 每个策略都需要更新的'每日数据'

        # # 2. 策略的主要内容
        # # ====================================================
        # # ====================================================

        #     # 1. 金叉: (昨天<0, 今天>0): 需要 buying it
        # if today_diff_score>0 and self.last_diff_score<=0:
        #     buying_count = self.cash_balance // self.today_price # [暂时设定] (待优化)
        #     self.buying(buying_price=self.today_price, buying_count=buying_count)
        #
        #     # 2. 死叉: (昨天>0, 今天<0): 需要 selling it
        # if today_diff_score<0 and self.last_diff_score>=0:
        #     selling_count = self.chicang_stock_count # 当出现死叉时, 把所有的持仓全部抛出
        #     self.selling(selling_price=self.today_price, selling_count=selling_count)

        # TODO: 止损稳赢策略
        pass

        # # ====================================================
        # # ====================================================


        # 3. 每晚:更新必要数据
        self.do_every_evening() # 每个策略都需要更新的'每日数据'





    def run(self, first_percentage=0.03, stoploss_rate=0.05, selected_days=100, start_date=None, end_date=None, pt=True):
        """
            function: 整个'策略类'的执行入口
            params:
                selected_avg: 选择哪两根均线作为长短线的计算差值
                selected_days:
                    1. 回测的天数 (越久远的数据可能与当前的市场差异很大, 而且也伴随通货膨胀之类的影响因素)
                    2. 日期的默认选取方式: 从今天开始, 倒退几天
                    3. 这里的天数: 是指实际天数 (不是"股市的交易天数")
                start_date/end_date: 日期的str标准格式: 2020-07-20
                pt: 是否需要打印输出

            return:
                self.asset_dict (整体的收益情况) (dict类型)

        """

        # 1. 初始化数据(清零)
            # (不清零的话, 会导致重复run方法时候出错)
        self.init_data(selected_days=selected_days, start_date=start_date, end_date=end_date, pt=pt)

        # A. 当存在'该时间段内的df'时:
        if len(self.hist_df) > 0:
            # self.selected_avg = selected_avg # 长短线指标选择 (几日均线与几日均线的差值)
            # self.all_arg_msg = f"【参数】:stock_code:{self.stock_code}; days:{self.selected_days} (start:{self.start_date} -- end:{self.end_date})"
            # if self.pt == True:
            #     print("\n\n", self.all_arg_msg, "\n")

            # 2. 执行策略计算
            self.hist_df.apply(self.strategy_cal, axis=1) # 运用上面的"策略计算"函数 (会把所有"买卖操作"数据存入 self.cash_flow_lst)

            # 3. 查看收益情况
            self.asset_dict = self.return_asset_dict() # 返回"账户资产相关信息"

            return self.asset_dict

        # B. 当不存在'该时间段内的df'时:
        else:
            return {} # 返回空dict


















def main():
    o = OrgModel(Zqdm="002505")
    d = o.run(30)

    o = DoubleMovingAverage("0.159995", init_cash_balance=100000)
    asset_dict = o.run(selected_days=30, selected_avg="5v10")
    print(asset_dict)







if __name__ == '__main__':
    print("Start test!\n\n")
    main()
    print("\n\n\nIt's over!")



"""
author : kerwin
function : personal settings
note : 用于修改所有文件的默认配置信息

警示： 以后再配置默认路径的时候，不要带上尾巴上的‘/’.  实际应用中看着不舒服。。。
        ("D:/自如/" 由于历史原因，不想改了。。。) //k200207:已经修改回来了
"""


"""
mac 和 linux 的路径除了开头的/Users 替换成 /root 外，其他是一样的。
"""



# 自动判断该系统是 ’windows' 还是 ‘mac’
import platform
import sys, os


# 配置信息
user_name   = "kerwin"




    # win
if platform.system() == "Windows":
    PYTHON = "python"
    HOST = "127.0.0.1"
    REMOTE_HOST = "121.40.240.153" # 阿里云的ip
    REMOTE_HOST = "kw618.xyz" # 阿里云的ip
    FILE_PATH_FOR_DESKTOP = "C:/Users/Administrator/Desktop"
    FILE_PATH_FOR_MYCODE = f"D:/{user_name}/MyCode"
    FILE_PATH_FOR_LIBS = f"D:/{user_name}/MyCode/Libs"
    FILE_PATH_FOR_ZIRU_CODE = f"D:/{user_name}/MyCode/BusinessProj/Ziru"
    FILE_PATH_FOR_ZIRU = "D:/自如"

    # mac
elif platform.system() == "Darwin":
    PYTHON = "python3"
    HOST = "127.0.0.1" # 本地ip  (mac上的代码默认用本地ip，若个别脚本有特别需要，可以设置成REMOTE_HOST)
    REMOTE_HOST = "121.40.240.153" # 阿里云的ip
    REMOTE_HOST = "kw618.xyz" # 阿里云的ip
    FILE_PATH_FOR_HOME = f"/Users/{user_name}" # 家目录
    FILE_PATH_FOR_DESKTOP = f"/Users/{user_name}/Desktop"
    FILE_PATH_FOR_MYBOX = f"/Users/{user_name}/MyBox"
    # FILE_PATH_FOR_KW618 = f"/Users/{user_name}/MyBox/kw618/kw618"  # 尽量减少使用这个文件路径 (因为它是pypi包, 不应该有固定路径!!!)
    FILE_PATH_FOR_MYCODE = f"/Users/{user_name}/MyBox/MyCode" # 最顶层代码目录
    FILE_PATH_FOR_LIBS = f"/Users/{user_name}/MyBox/MyCode/Libs" # 指: 顶层目录中用于存储 "资源" 的目录
    FILE_PATH_FOR_ZIRU_CODE = f"/Users/{user_name}/MyBox/MyCode/BusinessProj/Ziru" # 指:"自如项目代码"的路径
    FILE_PATH_FOR_ZIRU = f"/Users/{user_name}/MyBox/工作相关/自如" # 指:"自如excel, ppt等文件的目录"的路径
    FILE_PATH_FOR_VUE = f"/Users/{user_name}/MyBox/VueProj"


    # linux
elif platform.system() == "Linux":
    PYTHON = "python3"
    HOST = "127.0.0.1"
    REMOTE_HOST = "121.40.240.153" # 阿里云的ip
    REMOTE_HOST = "kw618.xyz" # 阿里云的ip
    FILE_PATH_FOR_HOME = f"/root" # 家目录
    FILE_PATH_FOR_DESKTOP = f"/root/{user_name}/Desktop"
    FILE_PATH_FOR_MYBOX = f"/root/{user_name}/MyBox"
    # FILE_PATH_FOR_KW618 = f"/root/{user_name}/MyBox/kw618/kw618" # 尽量减少使用这个文件路径 (因为它是pypi包, 不应该有固定路径!!!)
    FILE_PATH_FOR_MYCODE = f"/root/{user_name}/MyBox/MyCode"
    FILE_PATH_FOR_LIBS = f"/root/{user_name}/MyBox/MyCode/Libs"
    FILE_PATH_FOR_ZIRU_CODE = f"/root/{user_name}/MyBox/MyCode/BusinessProj/Ziru"
    FILE_PATH_FOR_ZIRU = f"/root/{user_name}/MyBox/工作相关/自如"
    FILE_PATH_FOR_VUE = f"/root/{user_name}/MyBox/VueProj"



# # 自动搜索kw618库包的位置 (暂时不用pypi的包, 所以可以定死"FILE_PATH_FOR_KW618"的路径) (上面已经定死, 这里后期优化)
site_pkgs_path = [path for path in sys.path if "packages" in path]
FILE_PATH_FOR_KW618 = ""
    # 1. 先看本地目录路径中是否有 'kw618'
if os.path.exists(f"{FILE_PATH_FOR_MYBOX}/kw618/kw618"):
    FILE_PATH_FOR_KW618 = f"{FILE_PATH_FOR_MYBOX}/kw618/kw618"
    kw618_pkg = "Local"
else:
    # 2. 再看第三方库中是否有 'kw618'
    for path in site_pkgs_path:
        if os.path.exists(f"{path}/kw618"):
            FILE_PATH_FOR_KW618 = f"{path}/kw618"
            kw618_pkg = "Third-party Libs"
            break
    # 3. 都没有, 则报警告
if not FILE_PATH_FOR_KW618:
        kw618_pkg = "There isn't kw618 pkg."
        print(
            "\n\n[kk警告]:  在本地 site-packages 路径中找不到kw618库, 尝试使用 pip3 install -i https://pypi.org/simple kw618  重新下载kw618库\n"
            "(使用pypi原地址而不是镜像, 才可以下载到最新上传的库包)\n\n"
        )


# 检查是否存在核心文件路径; (没有则抛出'警告'[不能是'报错', 否则体验更差])
error_path = []
for path in [FILE_PATH_FOR_DESKTOP, FILE_PATH_FOR_MYCODE]:
    if not os.path.exists(path):
        error_path.append(path)
if error_path:
    print(
        "[kk警告]:\n不存在以下文件路径:",
        "\n".join(error_path),
        "\n建议:"
        "\n  1. 修改user_name"
        "\n  2. 创建对应目录后, 重试"
        f"\n    (或在'{FILE_PATH_FOR_KW618}'目录下, 直接删除上面源代码)\n"
    )



if __name__ == "__main__":
    print("start!")
    print(FILE_PATH_FOR_DESKTOP)
    # print(FILE_PATH_FOR_KW618)
    print("end!")



import pandas as pd
import numpy as np
import math
import collections
import pymongo
import json
import copy
import hashlib
from io import StringIO

import warnings
warnings.filterwarnings("ignore")

# 导入常用的固定路径(多平台通用)
from kw618._file_path import *










class Transaction():

    # 每笔交易费用率
    # ==================
        # 1. 政府印花税: 万10
        # 2. 券商手续费: 万1.8不免5
        # 3. 规费: 万0.69
        # 4. 中登过户费(上海): 万0.2
    def __init__(self, commission_fee_rate=1.8, gui_fee_rate=0.69, transfer_fee_rate=0.2, stamp_tax_rate=10):
        # """
        # 印花税>手续费>规费>过户费
        # (很多券商的手续费是包含规费和过户费的, 这些是必须要交的, 所以手续费中实际能赚到的其实也不多)
        # (如果手续费率包含规费率, 那规费率记为0)
        # """
        u = 0.0001
        self.commission_fee_rate = commission_fee_rate * u
        self.gui_fee_rate = gui_fee_rate * u
        self.transfer_fee_rate = transfer_fee_rate * u
        self.stamp_tax_rate = stamp_tax_rate * u


    def cal_total_fee(self, transaction_amount=10000, transaction_type="in"):
        """
            params:
                transaction_amount: 交易金额
        """
        # 1. 券商手续费
        commission_fee = max(transaction_amount * self.commission_fee_rate, 5)
        # 2. 其他各种银行/政府费率
        other_fee = transaction_amount * (self.stamp_tax_rate + self.gui_fee_rate + self.transfer_fee_rate)
        # 总计费用
        # ==================
        total_fee = commission_fee + other_fee
        # 总费用占比
        # ==================
        fee_proportion = round(total_fee/transaction_amount, 5)
        print(f"\n该笔交易为: {transaction_amount}元 的金额, 共产生 {round(total_fee, 1)}元 的服务费, 占比 {str(fee_proportion*100)}%;\n")
        return total_fee



    def test(self):
        # 从1元到100w元
        for i in [1, 10, 100, 1000, 10000, 100000, 1000000]:
            self.cal_total_fee(i)





















#




# 导入常用的固定路径(多平台通用)
from kw618._file_path import *
from MyQR import myqr




words = "http://kw618.xyz" # 黑点密集多比较合适
words = "http://kw618.xyz/price/detail?house_id=HZZRGY0818218414_02"  # 黑点密集度有点偏高了

def make_qrcode(words, save_name="t1.jpg", save_dir=FILE_PATH_FOR_DESKTOP):
    "生成二维码; (如果http开头, 扫描后会自动跳转到该页面)"
    myqr.run(
    	words, # words 的字符数越多, 二维码越复杂, 黑点越多
        version=1, # version 的数值越大, 二维码越复杂, 黑点越多
        level='L', # 从L/M/Q/H顺序, 字母越靠右, 二维码越复杂, 黑点越多 (L是最简单的)
        picture=None,
        colorized=False,
        contrast=1.0,
        brightness=1.0,
        # save_name="kw_v1.jpg",
        save_name=save_name,
        save_dir=save_dir,
    )




# # 网上的案例
# # =================
# myqr.run(
#     words='http://www.baidu.com',
#     # 扫描二维码后，显示的内容，或是跳转的链接
#     version=5,  # 设置容错率
#     level='H',  # 控制纠错水平，范围是L、M、Q、H，从左到右依次升高
#     picture='test.gif',  # 图片所在目录，可以是动图
#     colorized=False,  # 黑白(False)还是彩色(True)
#     contrast=1.0,  # 用以调节图片的对比度，1.0 表示原始图片。默认为1.0。
#     brightness=1.0,  # 用来调节图片的亮度，用法同上。
#     save_name='test_qrcode2.gif',  # 控制输出文件名，格式可以是 .jpg， .png ，.bmp ，.gif
# )
# # 如果不设置这个save_name参数，并且也不想生成图片、动态二维码时，默认生成的图片名称为qrcode.png
# # 如果不设置这个save_name参数，并且加了picture参数，如test.gif，生成的即为test_qrcode.gif，如test.jpg，生成的即为test_qrcode.jpg
# # 报错：Wrong picture! Input a filename that exists and be tailed with one of {'.jpg', '.png', '.bmp', '.gif'}!
# # 解决方案：picture与save_name 设置一致












#



import time
import json
import redis
from kw618.k_requests.utils_requests import myRequest
from kw618.k_requests.my_cookies import PriceSysCookies

# 导入必要变量
    ## 1. redis
r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)
    ## 2. requests
req = myRequest().req


class RobustZiru():
    def __init__(self, user_name="lvzc1", user_passwd="Lzc15168201914*", AccessToken=""):
        """
            坑点:
                1. 使用该类的req方法访问ziru系统时, 第一次不会用到 user_name/user_passwd
                2. 这里的 user_name/user_passwd, 只有在第二次req请求时候才会被用到

        """
        self.user_name = user_name
        self.user_passwd = user_passwd
        self.AccessToken = AccessToken



    def req(self, url, req_method="get", data=None, other_headers={}):
        """
        此函数是在req的基础上做了定制化处理,
        考虑到多种可能的返回值,分流处理,稳定地得到最终需要的数据.

        function:
            - 专门针对自如价格相关系统的 robust_req
            - 情况1: 如果token仍然有效, 直接正常访问
            - 情况2: 当访问失败时(token过期), 会自动重新生成token, 并更新到本地redis缓存中

        更新日期: 2020-07-07
        """
        for retry_time in range(1, 4):
            try:
                # 1.获取token
                # (二者选其一, 一旦token无效, 就使用账号密码重新生成token)
                if self.AccessToken:
                    # token获取方式1: 直接传入
                    # 一定要给一个正确的token  (web上, 最好是直接给我一个能用的/有效的token)
                    token = self.AccessToken
                else:
                    # token获取方式2: 从redis中取出"已存在的token"
                    token = r.get("price_system:AccessToken")
                headers = {
                    "Access-Token":token,
                }
                other_headers.update(headers)
                # 2.发起req
                res_txt = req(url, other_headers=other_headers, req_method=req_method, data=data)
                res_json = json.loads(res_txt, strict=False)
                # 3.如果验证失败，重新获取cookies
                    # 40002:token过期;     40017:无数据权限;
                if res_json.get("code", 0) == "40002" or res_json.get("code", 0) == "40017":
                    print("token验证失败，使用post模拟登陆，获取cookies....")
                    # token获取方式3: 使用账号密码, 重新生成
                    token = PriceSysCookies(user_name=self.user_name, user_passwd=self.user_passwd).get_target_cookie()
                    r.set("price_system:AccessToken", token)
                    r.expire("price_system:AccessToken", 1800)
                    print("\n最新token:{}\n".format(token))
                    other_headers = {
                        "Access-Token":token,
                    }
                    res_txt = req(url, other_headers=other_headers, req_method=req_method, data=data)
                    res_json = json.loads(res_txt, strict=False)
                # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> (以上是关于cookies的验证)

                # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> (以下是关于response的分流处理)
                # 1..如果获取到的code字段=40001,说明 post方法缺少必要参数
                if res_json.get("code", 0) == "40001":
                    return {}, "miss" # post方法缺少必要参数
                # 2.如果获取到的status字段=5,说明发生其他问题,待处理
                elif res_json.get("status", 0) == "5": #此处占坑,作为其他错误response的捕获#
                    return {}, "failure"
                # 3.如果以上错误response都没捕获到,说明response正常,返回该正常的response
                else:
                    return res_json, "success" #"被期望"的正常情况#

            except Exception as e:
                print(e)
                ## 服务器确实只返回这样的数据,可能是被反爬了,也可能是发送给服务器的数据有问题
                ## (大概率是被反爬了:1.ip被封;2.cookies过期;3.headers/query缺失等)
                ## 需要在上面代码块重新调整"返回值的判断逻辑"!!
                print("response 非目标对象, 可能被反爬!")
                print("马上进行第 {} 次req重试\n[睡眠3秒]".format(retry_time))
                time.sleep(3)
                continue
            return {}, "error"



# 待删: 已经用"类"的形式实现了!
# def robust_req_ziru(url, req_method="get", data=None, other_headers={}, AccessToken=""):
#     """
#     此函数是在req的基础上做了定制化处理,
#     考虑到多种可能的返回值,分流处理,稳定地得到最终需要的数据.
#
#     function:
#         - 专门针对自如价格相关系统的 robust_req
#         - 情况1: 如果token仍然有效, 直接正常访问
#         - 情况2: 当访问失败时(token过期), 会自动重新生成token, 并更新到本地redis缓存中
#
#     更新日期: 2020-07-07
#     """
#     for retry_time in range(1, 4):
#         try:
#             # 1.获取token
#             if AccessToken:
#                 # 一定要给一个正确的token
#                 token = AccessToken
#             else:
#                 # 从redis中取出"已存在的token"
#                 token = r.get("price_system:AccessToken")
#             headers = {
#                 "Access-Token":token,
#             }
#             other_headers.update(headers)
#             # 2.发起req
#             res_txt = req(url, other_headers=other_headers, req_method=req_method, data=data)
#             res_json = json.loads(res_txt, strict=False)
#             # 3.如果验证失败，重新获取cookies
#             if res_json.get("code", 0) == "40002":
#                 print("token验证失败，使用post模拟登陆，获取cookies....")
#                 token = PriceSysCookies(user_name=user_name, user_passwd=user_passwd).get_target_cookie()
#                 r.set("price_system:AccessToken", token)
#                 r.expire("price_system:AccessToken", 1800)
#                 print("\n最新token:{}\n".format(token))
#                 other_headers = {
#                     "Access-Token":token,
#                 }
#                 res_txt = req(url, other_headers=other_headers, req_method=req_method, data=data)
#                 res_json = json.loads(res_txt, strict=False)
#             # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> (以上是关于cookies的验证)
#
#             # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> (以下是关于response的分流处理)
#             # 1..如果获取到的code字段=40001,说明 post方法缺少必要参数
#             if res_json.get("code", 0) == "40001":
#                 return {}, "miss" # post方法缺少必要参数
#             # 2.如果获取到的status字段=5,说明发生其他问题,待处理
#             elif res_json.get("status", 0) == "5": #此处占坑,作为其他错误response的捕获#
#                 return {}, "failure"
#             # 3.如果以上错误response都没捕获到,说明response正常,返回该正常的response
#             else:
#                 return res_json, "success" #"被期望"的正常情况#
#
#         except Exception as e:
#             print(e)
#             ## 服务器确实只返回这样的数据,可能是被反爬了,也可能是发送给服务器的数据有问题
#             ## (大概率是被反爬了:1.ip被封;2.cookies过期;3.headers/query缺失等)
#             ## 需要在上面代码块重新调整"返回值的判断逻辑"!!
#             print("response 非目标对象, 可能被反爬!")
#             print("马上进行第 {} 次req重试\n[睡眠3秒]".format(retry_time))
#             time.sleep(3)
#             continue
#         return {}, "error"


def test():
    # 测试: robust_req_ziru() 函数
    url = "http://price.ziroom.com/backend/adjustprice/getApprovalDedail?id=505180037330304"
    res_dict, my_status = robust_req_ziru(url)
    print(res_dict, my_status)







if __name__ == '__main__':
    print("start test!")
    test()
    print("\n\n\neverything is ok!")



import time
import redis
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

# 导入常用的固定路径(多平台通用)
from kw618._file_path import *
# 本脚本依赖很多 utils_requests的函数和模块, 直接用*  (注意要避免循环导入问题)
from kw618.k_requests.utils_requests import *

req = myRequest().req



class PriceSysCookies():
    name = "price_sys_cookies"

    # def __init__(self, user_name="lvzc1", user_passwd="Lzc15168201914*"):
    def __init__(self, user_name="yangjj4", user_passwd="Yangjiajing131#"):
        self.userEmail = user_name
        self.password = user_passwd
        # password = "OnHKU3EzPweR9sxjC42mXw=="
        # password = "OnHKU3EzPwcih0Hr+9bF5g=="
        self.encrypted_password = self.get_encrypted_password()

    def get_encrypted_password(self):
        "获取js加密后的用户密码"
        js_full_path = f"{FILE_PATH_FOR_KW618}/k_requests/js/encryption_4_price_system.js"
        encrypted_password = exec_js_function(js_full_path, "encryptByDES", self.password, "#z@i!r*o%o&m^")
        return encrypted_password


    def get_user_info(self):
        "获取登录后返回的用户信息: 包括名字/部门/职位/电话"
        url = "http://price.ziroom.com/backend/user/getAllInformation"
        data = {
            "userEmail" : self.userEmail,
            "password" : self.encrypted_password,
        }
        res_str = req(url=url, data=data, req_method="post")
        res_dic = json.loads(res_str)
        return res_dic


    def get_cookies(self):
        login_url = "http://price.ziroom.com/backend/user/login"

        password_urlencode = parse.quote(self.encrypted_password) # 进行url 编码
        data = "userEmail={}&password={}".format(self.userEmail, password_urlencode)
        other_headers = {
            "Connection":"keep-alive",
            "content-type":"application/x-www-form-urlencoded; charset=UTF-8;",
        }
        res_txt = req(login_url, req_method="post",
                        data=data, other_headers=other_headers)
        res_json = json.loads(res_txt, strict=False)
        cookies_dict = res_json # 说明是dict类型的cookies
        return cookies_dict



    def get_target_cookie(self):
        cookies_dict = self.get_cookies()
        TOKEN = cookies_dict.get("data", {}).get("accessToken","618")
        print("\n-------------------\n获取到最新的token：{0}\n\n".format(TOKEN))

        if TOKEN:
            return TOKEN
        else:
            return "error"

# token = PriceSysCookies(user_name="zhangh47", user_passwd="Wsxokn@04").get_target_cookie()
# token = PriceSysCookies(user_name="lvzc1", user_passwd="Lzc15168201914!").get_target_cookie()
# print(token)
# sys.exit()




#
# class PriceSysCookiesByS():
#     "千年不用的, 暂时注释, 需要时候再开启"
#     "ByS: 指用Selenium来进行爬取(与requests库的逻辑不一样)"
#     name = "price_sys_cookies_by_selenium"
#
#     def get_cookies(self):
#         try:
#             # browser = webdriver.PhantomJS()
#             browser = webdriver.Chrome()
#
#             browser.get("http://price.ziroom.com/auth.html")
#             username_input = browser.find_element_by_css_selector("#userAccount")
#             username_input.send_keys("lvzc1")
#
#             password_input = browser.find_element_by_css_selector("#password")
#             password_input.send_keys("Lzc15168201914!")
#
#             login_btn = browser.find_element_by_css_selector("#div-login > div > div:nth-child(4) > div > button")
#             login_btn.send_keys(Keys.ENTER)
#
#             wait = WebDriverWait(browser, 50)
#             wait.until(EC.presence_of_element_located((By.CLASS_NAME, "slimScrollDiv")))
#
#             # print(browser.current_url)
#             # 浏览器对象get_cookies()后,获取到的是lst类型,包含的每个dict都包含"name,value,domain,path"等
#             # browser_cookies_lst 为非标准的cookies形式
#             browser_cookies_lst = browser.get_cookies()
#             print(browser_cookies_lst)
#         finally:
#             print("done")
#             browser.close()
#         ## 输出返回的是 "非标准的"cookies
#         return browser_cookies_lst
#
#
#     def get_target_cookie(self):
#         browser_cookies_lst = self.get_cookies()
#         for cookie_dict in browser_cookies_lst:
#             if cookie_dict.get("name", 0) == "accessToken":
#                 TOKEN = cookie_dict.get("value")
#         print("\n-------------------\n获取到最新的token：{0}\n\n".format(TOKEN))
#         if TOKEN:
#             return TOKEN
#         else:
#             return "error"
#
#
#     # browser = webdriver.PhantomJS()
#     # 生成驱动对象是挺耗时的，所以如果要循环获得多个cookies时，不要马上close，等所有都获取完后，再关闭驱动对象！
#
#     # 打开多个浏览器,循环获取cookies  (基本上用不到这个函数...)
#     def browser_loop(self, times):
#         for i in range(times):
#             cookies = get_cookies()
#         return cookies





class NoahSysCookies_v2():

    def __init__(self, user_name, user_passwd):
        self.user_name = user_name
        self.user_passwd = user_passwd
        # self.user_name = "lvzc1"
        # self.user_passwd = "Lzc15168201914*"
        js_full_path = f"{FILE_PATH_FOR_KW618}/k_requests/js/encryption_4_ziroom_intranet.js"
        self.encrypted_password = exec_js_function(js_full_path, "encodeAes", self.user_passwd)
        # self.ziroom_userName = "lvzc1"
        # self.ziroom_token = "9c0b15d0758030f5f64d419eac77fd47"
        self.req_obj = myRequest()
        self.req = self.req_obj.req


    def get_jsessionid(self):
        url = "http://cas.ziroom.com/CAS/login"
        res_obj = self.req(url, is_obj=True, use_session=True)
        res_selector = Selector(text=res_obj.text)
        # 能够实现把cookiejar对象转化为字典
        ### 获取JSESSIONID
        cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
        print(cookies)
        jsessionid = cookies.get("JSESSIONID", "")
        self.jsessionid = jsessionid
        self.lt = res_selector.xpath('//*[@id="loginForm"]/input[1]/@value').extract_first()
        self.execution = res_selector.xpath('//*[@id="loginForm"]/input[2]/@value').extract_first()
        self._eventId = res_selector.xpath('//*[@id="loginForm"]/input[3]/@value').extract_first()
        return jsessionid


    def get_castgc(self, authCode=""):
        if "jsessionid" not in dir(self):
            self.get_jsessionid()
        jsessionid = self.jsessionid
        url = "http://cas.ziroom.com/CAS/login;jsessionid={}".format(jsessionid)
        other_headers = {
            "Cookie":"JSESSIONID={jsessionid}".format(jsessionid=jsessionid),
            "Content-Length": "148",
            "Cache-Control": "max-age=0",
            "Origin": "http://cas.ziroom.com",
            "Upgrade-Insecure-Requests": "1",
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3",
            "Referer": "http://cas.ziroom.com/CAS/login",
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,zh-TW;q=0.6",
            "Connection": "keep-alive"
            }
        data = "username={}&password={}&authCode={}&lt={}&execution={}&_eventId={}"
        username = self.user_name
        # password = "Vhydw+8N3+dcg5DxFvuwWg==" # 这部分的加密，已经可以用线程的js代码自动加密生成！！（已破解）
        password = parse.quote(self.encrypted_password)
        if not authCode:
            authCode = input("请手动输入动态密码:\n")
        lt = self.lt
        execution = self.execution
        _eventId = self._eventId
        data = data.format(username, password, authCode, lt, execution, _eventId)

        res_obj = self.req(url=url, other_headers=other_headers, data=data, is_obj=True, req_method="post", use_session=True)
        cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
        print(cookies)
        castgc = cookies.get("CASTGC", "")
        self.castgc = castgc
        return castgc


    def get_new_jsessionid(self):
        url = "http://z.ziroom.com/"
        res_obj = self.req(url=url, is_obj=True, use_session=True)
        cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
        new_jsessionid = cookies.get("JSESSIONID")
        print("new_jsessionid:", new_jsessionid)
        self.new_jsessionid = new_jsessionid
        return new_jsessionid


    def req_security_index(self):
        url = "http://z.ziroom.com/security/security!index.action"
        other_headers = {"Cookie":"JSESSIONID={}".format(self.new_jsessionid)}

        res_obj = self.req(url=url, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l1 = res_obj.headers.get("location", "")
        self.location_l1 = location_l1

        # 2. 开始下一个跳转
        other_headers = {"Cookie":"JSESSIONID={};CASTGC={}".format(self.jsessionid, self.castgc)}
        res_obj = self.req(url=location_l1, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l2 = res_obj.headers.get("location", "")
        self.location_l2 = location_l2
        ticket = re.findall(r"ticket=(.*?com)", location_l2)
        ticket = ticket[0] if len(ticket) else ""
        self.ticket = ticket

        # 3. 开始下一个跳转  (使用ticket访问security！index.action)
        other_headers = {
            "Cookie":"JSESSIONID={}".format(self.new_jsessionid),
            "Upgrade-Insecure-Requests": "1",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "Referer": "http://z.ziroom.com/",
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,zh-TW;q=0.6",
            "Connection": "keep-alive",
        }
        url_l2 = "http://z.ziroom.com/security/security!index.action?ticket={}".format(self.ticket)
        res_obj = self.req(url=url_l2, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l3 = res_obj.headers.get("location", "")
        self.location_l3 = location_l3  # 正常来说，这个location又变成了最初的 security!index.action


        # 4. 开始下一个跳转
        other_headers = {"Cookie":"JSESSIONID={}".format(self.new_jsessionid)}
        res_obj = self.req(url=location_l3, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l4 = res_obj.headers.get("location", "")
        self.location_l4 = location_l4

        ##  获取其中的ziroom_token
        cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
        print(cookies)
        ziroom_token = cookies.get("ziroom_token")
        ziroom_userName = cookies.get("ziroom_userName")
        self.ziroom_token = ziroom_token
        self.ziroom_userName = ziroom_userName
        return ziroom_token



    def get_noah_session(self):
        url = "http://noah.ziroom.com/main?sys=inv"
        self.ziroom_userName = self.user_name
        self.ziroom_token = self.ziroom_token
        other_headers = {
            "Cookie":"ziroom_userName={ziroom_userName};ziroom_token={ziroom_token}".format(
            ziroom_userName=self.ziroom_userName, ziroom_token=self.ziroom_token
            )
        }


        res_obj = self.req(url=url, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
        noah_session = cookies.get("SESSION", "")
        print("noah_session:{}".format(noah_session))
        self.noah_session = noah_session
        location_l1 = res_obj.headers.get("location", "")
        self.location_l1 = location_l1

        # 2. 开始下一个跳转
        other_headers = {"Cookie":"JSESSIONID={};CASTGC={};ziroom_userName={};ziroom_token".format(
                            self.jsessionid, self.castgc, self.ziroom_userName, self.ziroom_token)
                        }
        res_obj = self.req(url=location_l1, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l2 = res_obj.headers.get("location", "")
        self.location_l2 = location_l2
        noah_ticket = re.findall(r"ticket=(.*?com)", location_l2)
        noah_ticket = noah_ticket[0] if len(noah_ticket) else ""
        self.noah_ticket = noah_ticket

        # 3. 开始下一个跳转  (使用ticket访问noah.ziroom.com)
        other_headers = {
            "Cookie":"ziroom_userName={};ziroom_token={};SESSION={}".format(
                self.ziroom_userName, self.ziroom_token, self.noah_session),
        }
        url_l2 = "http://noah.ziroom.com/main?sys=inv&ticket={}".format(self.noah_ticket)
        res_obj = self.req(url=url_l2, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l3 = res_obj.headers.get("location", "")
        self.location_l3 = location_l3  # 正常来说，这个location又变成了最初的 http://noah.ziroom.com/main?sys=inv


        # 4. 开始下一个跳转
        ### other_headers 沿用上面的headers
        res_obj = self.req(url=location_l3, is_obj=True, other_headers=other_headers, allow_redirects=False, use_session=True)
        location_l4 = res_obj.headers.get("location", "")
        cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
        noah_id = cookies.get("noah_id")
        self.noah_id = noah_id

        return noah_session


    # 主要的获取流程
    def run(self, authCode=""):
        # 0. 先获取redis的连接对象
        r = redis.StrictRedis(host="localhost", port=6379, db=0, decode_responses=True)

        if not authCode:
            authCode = input("请您输入内网验证码：")
        self.get_castgc(authCode=authCode)
        self.get_new_jsessionid()
        self.req_security_index()
        noah_session = self.get_noah_session()
        r.set("noah_session", noah_session)
        # 设置30分钟过期时间
        r.expire("noah_session", 1800)
        return noah_session
# o = NoahSysCookies_v2()
# o.run(authCode="222590")
# sys.exit()


#
# class NoahSysCookies():
#     "上面已经有第二个版本的实现了, 弃之, 待删"
#
#     def __init__(self):
#         self.ziroom_userName = "lvzc1"
#         self.ziroom_token = "9c0b15d0758030f5f64d419eac77fd47"
#
#     def get_jsessionid(self):
#         url = "http://cas.ziroom.com/CAS/login"
#         res_obj = req(url, is_obj=True)
#         res_selector = Selector(text=res_obj.text)
#         # 能够实现把cookiejar对象转化为字典
#         ### 获取JSESSIONID
#         cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
#         print(cookies)
#         jsessionid = cookies.get("JSESSIONID", "")
#         self.jsessionid = jsessionid
#         self.lt = res_selector.xpath('//*[@id="loginForm"]/input[1]/@value').extract_first()
#         self.execution = res_selector.xpath('//*[@id="loginForm"]/input[2]/@value').extract_first()
#         self._eventId = res_selector.xpath('//*[@id="loginForm"]/input[3]/@value').extract_first()
#         return jsessionid
#
#
#     def get_castgc(self, authCode=""):
#         if "jsessionid" not in dir(self):
#             self.get_jsessionid()
#         jsessionid = self.jsessionid
#         url = "http://cas.ziroom.com/CAS/login"
#         other_headers = {
#             "Cookie":"JSESSIONID={jsessionid}".format(jsessionid=jsessionid),
#             "Content-Length": "148",
#             "Cache-Control": "max-age=0",
#             "Origin": "http://cas.ziroom.com",
#             "Upgrade-Insecure-Requests": "1",
#             "Content-Type": "application/x-www-form-urlencoded",
#             "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3",
#             "Referer": "http://cas.ziroom.com/CAS/login",
#             "Accept-Encoding": "gzip, deflate",
#             "Accept-Language": "en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,zh-TW;q=0.6",
#             "Connection": "keep-alive"
#             }
#         data = "username={}&password={}&authCode={}&lt={}&execution={}&_eventId={}"
#         username = "lvzc1"
#         password = "Vhydw+8N3+dcg5DxFvuwWg==" # 这部分的加密，已经可以用线程的js代码自动加密生成！！（已破解）
#         password = parse.quote(password)
#         if not authCode:
#             authCode = input("请手动输入动态密码:\n")
#         lt = self.lt
#         execution = self.execution
#         _eventId = self._eventId
#         data = data.format(username, password, authCode, lt, execution, _eventId)
#
#         res_obj = req(url=url, other_headers=other_headers, data=data, is_obj=True, req_method="post")
#         cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
#         print(cookies)
#         castgc = cookies.get("CASTGC", "")
#         self.castgc = castgc
#         return castgc
#
#
#
#     def get_unverified_session(self):
#         url = "http://noah.ziroom.com/main?sys=inv"
#         # self.ziroom_userName = "lvzc1"
#         # self.ziroom_token = "9b8199276caa50c1528a1d53d170e0b1"
#         other_headers = {
#             "Cookie":"ziroom_userName={ziroom_userName};ziroom_token={ziroom_token}".format(
#             ziroom_userName=self.ziroom_userName, ziroom_token=self.ziroom_token
#             )
#         }
#         res_obj = req(url=url, is_obj=True, other_headers=other_headers, allow_redirects=False)
#         cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
#         unverified_session = cookies.get("SESSION", "")
#         print("unverified_session:{}".format(unverified_session))
#         self.unverified_session = unverified_session
#
#         # 上面的allow_redirects参数设为False后，下面需要手动跳转到location的地址
#         location_l1 = res_obj.headers.get("location", "")
#         self.location_l1 = location_l1
#         print(location_l1)
#         return unverified_session
#
#
#
#
#     # 实际上就是获取那个location_l2即可！！（ticket是locaiton中的一个参数）
#     def get_ticket(self, authCode=""):
#         if "castgc" not in dir(self):
#             self.get_castgc(authCode=authCode)
#         if "unverified_session" not in dir(self):
#             self.get_unverified_session()
#         url = self.location_l1
#         other_headers = {
#             "Cookie":"JSESSIONID={jsessionid};CASTGC={castgc};ziroom_userName={ziroom_userName};ziroom_token={ziroom_token}".format(
#             jsessionid=self.jsessionid, castgc=self.castgc, ziroom_userName=self.ziroom_userName, ziroom_token=self.ziroom_token
#             )
#         }
#         res_obj = req(url=url, is_obj=True, other_headers=other_headers, allow_redirects=False)
#         cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
#         # 上面的allow_redirects参数设为False后，下面需要手动跳转到location_l2的地址
#         location_l2 = res_obj.headers.get("location", "")
#         print(location_l2)
#         ticket = re.findall(r"ticket=(.*?com)", location_l2)
#         ticket = ticket[0] if len(ticket) else ""
#         # print(ticket)
#         self.location_l2 = location_l2
#         self.ticket = ticket
#         return ticket
#
#
#     def get_verified_session(self, authCode=""):
#         if "location_l2" not in dir(self):
#             self.get_ticket(authCode=authCode)
#         if "unverified_session" not in dir(self):
#             self.get_unverified_session()
#         location_l2 = self.location_l2
#         url = location_l2
#         other_headers = {
#             "Cookie":"SESSION={session};ziroom_userName={ziroom_userName};ziroom_token={ziroom_token}".format(
#             session=self.unverified_session, ziroom_userName=self.ziroom_userName, ziroom_token=self.ziroom_token
#             )
#         }
#         res_obj = req(url, is_obj=True, allow_redirects=False, other_headers=other_headers)
#         # cookies = requests.utils.dict_from_cookiejar(res_obj.cookies)
#         # verified_session = cookies.get("SESSION")
#         # self.verified_session = verified_session
#
#         verified_session = self.unverified_session
#         print("SESSION:{}".format(verified_session))
#         return verified_session








def main():

    # cookies_1 = PriceSysCookiesByS().get_target_cookie()
    # print("cookies_1:\n{0}\n\n".format(cookies_1))

    cookies_2 = PriceSysCookies().get_target_cookie()
    print("cookies_2:\n{0}\n\n".format(cookies_2))



if __name__ == '__main__':
    print("start!")
    main()
    print("end!")



import pymysql
from sqlalchemy import create_engine






def convert_lst_2_str_tuple(lst):
    changed_lst = []
    for i in lst:
        if isinstance(i, int):
            # 如果i是int型, 需要转成str才行
            i_ = str(i)
            changed_lst.append(i_)
        elif isinstance(i, str):
            # 如果i是str型, 需要在字符串外面多添加一个''
            i_ = "'{}'".format(i)
            changed_lst.append(i_)
    s = ",".join(changed_lst)
    return "(" + s + ")"








def main():
    # pass
    print(convert_lst_2_str_tuple([2, 3, "ss", 3, "mm"]))


if __name__ == '__main__':
    print("start test!")
    main()
    print("\n\n\neverything is ok!")



"""
    因为kw618的init中只能导入全局变量/函数/类, 而无法导入类中的函数.
    所以, 其实把该模块作为一个"大的类", 里面都是类中实现某些功能的函数
    所以, docs_2_df 函数, 其实没必要归纳到类中. 这样显得层级很复杂, 而且也不方便外部脚本调用该函数.
"""
import pandas as pd
import numpy as np
import math
import collections
import pymongo
import json
import copy
import hashlib
from io import StringIO

import warnings
warnings.filterwarnings("ignore")

# 导入常用的固定路径(多平台通用)
from kw618._file_path import *

def import_data(
    in_file_name="in", end_index=None, field=None, is_df=True,
    in_file_path=None, encoding="gb18030", index_col=None,
    ):
    """
    in:csv文件
    out:df类型/类mongo类型
    function:  csv → df/mongo (默认转出:类mongo)

    notes: in_file_path 的优先级比 in_file_name 高。

    ttt:0214
    """
    if in_file_path:
        df = pd.read_csv(in_file_path, encoding=encoding, engine='python', index_col=index_col)
    else:
        df = pd.read_csv(FILE_PATH_FOR_DESKTOP+"/{0}.csv".format(in_file_name), encoding=encoding, engine='python', index_col=index_col)
    if is_df:
        return df
    # 1.需要返回的是某个字段的lst格式
    if field:
        field_lst = df[field].values[:end_index] # 得到的是np.array格式
        return list(field_lst) # 用list数据格式来返回
    # 2.返回的是mongo支持的docs
    df = df[:end_index]
    docs = df.T.to_dict().values()
    return docs



    #  也可以用于 "mongo → df"
def output_data(
    in_obj, out_file_name="out", ordered_field_lst=None,
    out_file_path=None, output=True, index=False, encoding="gb18030", export_excel=False,
    ):
    """
    in:类mongo/df
    out:csv文件
    function:  1.mongo/df  → csv
               2.mongo → df (这样output设为False即可)

    in_obj:    不管是mongo还是df,自动先转化成df,再用它来转csv

    tips: 如果需要 "mongo → df": output设置为False即可!
    notes: out_file_path 的优先级比 out_file_name 高。

    """

    # 1. 如果是 "类mongo" 类型, 先转化成df
    if isinstance(in_obj, pymongo.cursor.Cursor):
        # total_items = []
        # for doc in in_obj:
        #     # items = {i:str(j).strip() for i, j in zip(list(doc.keys()), list(doc.values()))}
        #     # 以下会按照mongo中存着的顺序进行输出!
        #     items = collections.OrderedDict({i:str(j).strip() for i, j in zip(list(doc.keys()), list(doc.values()))})
        #     total_items.append(items)
        # df = pd.DataFrame(total_items)
        df = pd.DataFrame(list(in_obj))  # 如果in_obj的数据量是上百万条, 其实这个操作很危险的!!
    elif isinstance(in_obj, pd.core.frame.DataFrame):
        df = in_obj

    # 2.确定字段的呈现顺序
    if ordered_field_lst:
        # 如果指定的df字段在df中并不存在,则把该字段remove掉.确保不报错
        for field in ordered_field_lst.copy():
            if field not in df.columns:
                print("字段 {} 不在df中,将其抛弃!".format(field))
                ordered_field_lst.remove(field)
        df = df[ordered_field_lst]  # 指定顺序

    # 3.看是否需要导出csv文件,如果不需要,直接返回df
    if not output:
        return df

    # 4. 最后,将df数据转成csv文件输出
    try:
        if out_file_path:
            if not export_excel:
                df.to_csv(out_file_path, index=index, encoding=encoding)
            else:
                df.to_excel(out_file_path, index=index, encoding=encoding)
        else:
            if not export_excel:
                df.to_csv(FILE_PATH_FOR_DESKTOP+"/{0}.csv".format(out_file_name), index=index, encoding=encoding)
            else:
                df.to_excel(FILE_PATH_FOR_DESKTOP+"/{0}.xlsx".format(out_file_name), index=index, encoding=encoding)
    except Exception as e:
        print(e)
        out_file_name = input("输出文件名出错,请重新键入文件名: ")
        df.to_csv(FILE_PATH_FOR_DESKTOP+"/{0}.csv".format(out_file_name), index=index, encoding=encoding)

    return df


# class KwPd():
#     def __init__(self):
#         pass
#
#     def docs_2_df(self, docs, ordered_field_lst=None):
#         """
#         把mongo的数据转化成df
#         """
#         df = output_data(docs, output=False, ordered_field_lst=ordered_field_lst)
#         return df



def docs_to_df(docs, ordered_field_lst=None):
    """
    把mongo的数据转化成df
    """
    df = output_data(docs, output=False, ordered_field_lst=ordered_field_lst)
    return df


def df_2_mongo(df):
    return df.T.to_dict().values() # 即：docs
def df_to_docs(df, is_lst=False):
    """
        notices:
            1. 这里传入的df的index, 应该只允许 0/1/2...999 的自然数.
                (不确定. 我把datetime对象作为index是会报错的)
            [巨坑]: 一定要注意 pd.concat([df1, df2], axis=0)的情况, 一定要加上 ignore_index=True !!!

            2. //20200812更新: 可以使用pandas自带的方法实现, 方便高效!!!
                (而且这种方式都不用担心出现上面的 判断pd.concat()中的索引重复导致转化缺失的问题)
    """
    # //20200812更新: 可以使用pandas自带的方法实现, 方便高效!!!
    # if is_lst:
    #     return list(df.T.to_dict().values())
    # else:
    #     return df.T.to_dict().values() # 即：docs

    docs = df.to_dict("records") # 高效!!
    return docs


def read_excel(in_file_name="in", in_file_path=None, sheet_name=None, need_to_concat=True):
    """
        params:
            sheet_name:
                传入None: 返回一个有序字典 OrderedDict([("<sheet名字>", <df对象>)])
                        ( 需要用sheet名来按键取值)
            need_to_concat:
                当没有指定"sheet_name"时, 默认把所有sheet合并, 返回合并后的df
                    (当need_to_concat为False时, 不自动合并sheet, 而是返回一个 'excel字典对象')
    """
    # 1. 先读取整个excel文件
    if in_file_path is not None:
        ordered_d = pd.read_excel(in_file_path, sheet_name=None)
    elif in_file_path is None:
        ordered_d = pd.read_excel(f"{FILE_PATH_FOR_DESKTOP}/{in_file_name}.xlsx", sheet_name=None)

    # 2. 读取对应sheet_name (返回df)
    if sheet_name != None:
        df = ordered_d.get(sheet_name)
        del ordered_d # 释放中间过程对象的内存
        return df
    # 3. 合并多个sheet, 返回合并后的df
    elif need_to_concat == True:
        concat_df = pd.concat([sheet for sheet in ordered_d.values()], axis=0, ignore_index=True)
        del ordered_d # 释放中间过程对象的内存
        return concat_df

    # 4. 返回这个excel字典对象 (每个键值对中, 以sheet的名字作为"键", 对应的df对象作为"值")
    return ordered_d


def sort_df(df, ordered_field_lst):
    # 1. 如果指定的字段在df中并不存在,则把该字段remove掉.确保不报错
    ordered_field_lst_copy = ordered_field_lst.copy()
    for field in ordered_field_lst_copy:
        if field not in df.columns:
            print("字段 {} 不在df中, 将其抛弃!".format(field))
            ordered_field_lst.remove(field)

    # 2. 把所需要保留的 "有序字段list" 作用在df上
    return df[ordered_field_lst]  # 指定顺序




# stackoverflow 白嫖来的函数，hhh
def read_mongo(collection_obj, query={}, need_to_show_dict={}, df_name="foo", need_to_convert_date=True):
    """
        params:
            need_to_convert_date: 是否需要在读取mongo数据的时候, 转化日期格式
        note: 白嫖来的函数, hhh

    """

    # 不需要获取"_id"字段
    need_to_show_dict.update({"_id":0})

    # Make a query to the specific DB and Collection
    # print(query, need_to_show_dict)
    cursor = collection_obj.find(query, need_to_show_dict)

    # Expand the cursor and construct the DataFrame
    df =  pd.DataFrame(list(cursor))

    df.df_name = df_name


    if ("crawl_date" in df.columns) and (need_to_convert_date==True):
        df["crawl_date"] = pd.to_datetime(df["crawl_date"])
    if ("date" in df.columns) and (need_to_convert_date==True):
        df["date"] = pd.to_datetime(df["date"])

    return df





def date_to_obj(date_str="today"):
    return pd.to_datetime(date_str)

def obj_to_date(date_obj, format_="%Y-%m-%d"):
    return date_obj.strftime(format_)

def get_today_obj(date_str="today"):
    return pd.to_datetime(date_str)

def get_days(days_str="1 d"):
    "1天的时间段"
    return pd.to_timedelta(days_str)

def get_this_time():
    "此刻的时间"
    today_obj = pd.to_datetime("today")
    return today_obj.strftime("%X") # 转成“15:43:36”这样的str形式

def get_this_date_time():
    "此刻的时间"
    today_obj = pd.to_datetime("today")
    return today_obj.strftime("%Y-%m-%d %X") # 转成 '2020-07-29 14:13:30' 这样的str形式


def get_today_date(date_str="today", format_="%Y-%m-%d"):
    "今天的日期"
    today_obj = pd.to_datetime(date_str)
    return today_obj.strftime(format_) # 转成“2019-02-28”这样的str形式
# 简化版的当前日期
def get_sim_today_date(date_str="today"):
    return get_today_date(date_str=date_str ,format_="%m%d") # 转成“0228”这样的str形式
# 简化版的昨天日期
def get_sim_yesterday_date(date_str="today"):
    return get_yesterday_date(date_str=date_str, format_="%m%d") # 转成“0228”这样的str形式

def get_sim_this_time():
    this_time_str = get_this_time() # '03:05:49'
    return this_time_str.replace(":", "") # '030549'

def get_yesterday_obj(date_str="today"):
    today_obj = pd.to_datetime(date_str)
    yesterday_obj = today_obj - pd.to_timedelta("1 d")
    return yesterday_obj

def get_yesterday_date(date_str="today", format_="%Y-%m-%d"):
    " 昨天的日期"
    today_obj = pd.to_datetime(date_str)
    yesterday_obj = today_obj - pd.to_timedelta("1 d")
    yesterday_date = yesterday_obj.strftime(format_) # 转成“20190228”这样的str形式
    return yesterday_date

def get_previous_date(date_str="today", days_str="10 d", format_="%Y-%m-%d"):
    today_obj = pd.to_datetime(date_str)
    days_obj = get_days(days_str)
    previous_date_obj = today_obj - days_obj
    previous_date = previous_date_obj.strftime(format_)
    return previous_date

def get_later_date(date_str="today", days_str="10 d", format_="%Y-%m-%d"):
    today_obj = pd.to_datetime(date_str)
    days_obj = get_days(days_str)
    previous_date_obj = today_obj + days_obj
    previous_date = previous_date_obj.strftime(format_)
    return previous_date



def get_this_month_first_date(date_str="today"):
    " 本月第一天的日期"
    today_obj = get_today_obj(date_str)
    this_month_first_obj = get_today_obj(today_obj.strftime("%Y-%m"))
    this_month_first_date = obj_to_date(this_month_first_obj)
    return this_month_first_date


def get_delta_days(start_date=get_yesterday_date(), end_date=get_today_date()):
    start_date_obj = get_today_obj(date_str=start_date)
    end_date_obj = get_today_obj(date_str=end_date)
    delta_days = (end_date_obj - start_date_obj).days
    return delta_days


def get_period_df(start_date=None, end_date=None, is_crawl_date=False):
    " 获取一段时间内的 <日期扩充表> "
    if start_date is None:
        this_month_first_date = get_this_month_first_date()
        start_date = this_month_first_date
    if end_date is None:
        end_date = get_today_date()

    # 两种方式截取 "日期范围"
    datetime_index = pd.date_range(start_date, end_date, freq="1d")
    if is_crawl_date: # 用"crawl_date"来选择 "日期范围"
        df = pd.DataFrame({"crawl_date":datetime_index})
        df["true_date"] = df.crawl_date - get_days("1 d")
    else: # 用"true_date"来选择 "日期范围"
        df = pd.DataFrame({"true_date":datetime_index})
        df["crawl_date"] = df.true_date + get_days("1 d")

    # 生成4中 str格式的日期  (用于后期透视)
    df["日期"] = df.true_date.dt.strftime("%Y-%m-%d")
    df["日期-年"] = df.true_date.dt.strftime("%Y") # series类型正常来说是不能直接strftime成str类型的, 必须要用.dt 方法才行
    df["日期-月"] = df.true_date.dt.strftime("%Y-%m")
    # 计算"日期-周"这个 '周度日期 '
    weekly_date_lst = []
    for count, date_str in enumerate(df["日期"][-1::-1]): # 对'日期'的series逆序
        if count % 7 == 0:
            tmp = date_str
        weekly_date_lst.append(tmp)
    df["日期-周"] = weekly_date_lst[-1::-1] # 上面逆序了, 现在逆序回来
    df["sim_true_date"] = df.true_date.dt.strftime("%m%d")
    df["sim_crawl_date"] = df.crawl_date.dt.strftime("%m%d")

    return df





def output_excel(df_lst, out_file_name="out", out_file_path=None, sheet_name_lst=None):
    from pandas import ExcelWriter
    if out_file_path is None:
        # 如果没有out_file_path: 默认放在桌面
        out_file_path = f"{FILE_PATH_FOR_DESKTOP}/{out_file_name}.xlsx"
    with ExcelWriter(out_file_path) as writer:
        for i, df in enumerate(df_lst):
            if sheet_name_lst:
                sheet_name = sheet_name_lst[i]
            else:
                sheet_name = f"sheet_{i}"
            df.to_excel(writer, sheet_name, index=False)
        writer.save()















def avg(lst):
    if isinstance(lst, list):
        if len(lst) <1:
            # raise myError("元素小于1!")
            return 0
    elif isinstance(lst, type(pd.Series())):
        if lst.size <1:
            # raise myError("元素小于1!")
            return 0
    sum = 0
    for count, e in enumerate(lst):
        # print(count, e)
        sum += int(float(e))
    lst_avg = sum/(count+1)
    # print(lst_avg)
    return int(lst_avg)


def merge_df(
    x_name, y_name, out_file_name="out",
    is_df=None, join_field="house_id", output=True):
    """
    function: 不仅可以合并df/csv, 还附带输出csv的功能
    """
    print(">>>1")
    if not is_df:
        # 如果 不是df， 就把这个当做文件名，导入
        x_df = import_data(x_name, is_df=True)
        y_df = import_data(y_name, is_df=True)
    else:
        # 如果 是df， 就直接把传入的x、y当做 df对象来使用
        x_df = x_name
        y_df = y_name
    print(">>>2")
    # pd.merge() 返回的不是df类型，而是function类型。 但这个function可以使用to_csv导出文件
    #  ??????   什么情况？ 之前测试的时候返回的不是df对象，现在测试发现又确实是df对象了。。。见鬼！
    merged_df = pd.merge(x_df, y_df, how="left", on=join_field)
    if not output:
        return merged_df
    print(">>>3")
    merged_df.to_csv(FILE_PATH_FOR_DESKTOP+"/{0}.csv".format(out_file_name), index=False, encoding="gb18030")
    print("合并成功!")

# merge_df("aaa", "bbb", out_file_name="zzzz")
# exit()


# def k_top(lst, top=1):
#     if isinstance(lst, list):
#         if len(lst) <1:
#             # raise myError("元素小于1!")
#             return 0
#     elif isinstance(lst, type(pd.Series())):
#         if lst.size <1:
#             # raise myError("元素小于1!")
#             return 0
#
#     lst = sorted(lst)
#     return lst[top-1]


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for np types """
    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)










def k_divide(lst, piece=5):
    """
    function: 按lst从小到大的顺序, 等分成piece份 小lst 返回
    return: 返回等分节点的lst (即:按照这几个值去截取, 就是5等分了)
    """
    if isinstance(lst, list):
        if len(lst) <1:
            # raise myError("元素小于1!")
            return 0
    elif isinstance(lst, type(pd.Series())):
        if lst.size <1:
            # raise myError("元素小于1!")
            return 0

    lst = sorted(lst)
    # 1. 打印原lst
    print(lst)
    node_order_lst = []
    node_lst = []
    for count in range(1, piece):
        node_order_value = round(len(lst) * (1/piece) * count) - 1 # 减一别忘了 (另外,这里返回的是顺序值,不是真实值)
        node_order_lst.append(node_order_value)
        node_lst.append(lst[node_order_value])
    # 2. 打印分好piece后的, 节点的顺序
    print(node_order_lst) # 是顺序
    print("值的lst: {}".format(node_lst)) # 是值

    piece_dict = {}
    count = 0
    while True:
        if count == piece:
            break
        elif count == 0:
            piece_dict.update({count+1 : lst[ : node_order_lst[count]+1]})
        elif count == piece-1:
            piece_dict.update({count+1 : lst[node_order_lst[count-1]+1 : ]})
        else:
            piece_dict.update({count+1 : lst[node_order_lst[count-1]+1 : node_order_lst[count]+1]})
        count += 1
    # 3. 打印根据上面的顺序, piece等拆分了lst后的dict
    print(piece_dict)
    return node_lst
    # return piece_dict


    # piece_lst = [] count = 0
    # while True:
    #     if count == piece:
    #         break
    #     elif count == 0:
    #         piece_lst.append(lst[ : node_order_lst[count]+1])
    #     elif count == piece-1:
    #         piece_lst.append(lst[node_order_lst[count-1]+1 : ])
    #     else:
    #         piece_lst.append(lst[node_order_lst[count-1]+1 : node_order_lst[count]+1])
    #     count += 1
    # # 3. 打印根据上面的顺序, piece等拆分了lst后的lst
    # print(piece_lst)
    # return piece_lst

# k_divide([3, 4, 5, 7, 2, 4, 46, 6, 7, 84, 4,5], 5)




def is_notnan_numeric(x):
    """
        numeric: 指所有数值: int/float (包括np.nan) (不包括None) (不包括'可以转成float的str')
        notnan_numeric: 指所有'非nan'的数值: int/float (不包括np.nan, 不包括None)
    """
    # 1. 若是数据集, 则直接返回False
    if isinstance(x, list) or isinstance(x, dict) or isinstance(x, set) or isinstance(x, tuple) or isinstance(x, np.ndarray):
        return False
    # 2. 是否为 None
    if x is None:
        return False
    # 3. 是否为 np.nan
    elif pd.isnull(x):
        return False
    # 4. 是否为 str
    elif isinstance(x, str):
        ### 注意: 这里防止x为可以被转成float的str, 先对str类型单独处理 (避免下一步造成错误判断)
        return False
    else:
        try:
            # 4. 如果可以被float()转化成float, 则x是为数值型, 返回True
            return isinstance(float(x), float)
        except:
            # 5. 不能转化, 则说明不是数值型
            return False


def safely_to_int(x, need_to_print=False):
    """
        save: 表示可以'安全'转化成'int'. 如果x为不能转化成int的数据, 则保留原样
        notice: 该函数是以'四舍五入'的方式转成int
    """
    # 如果是'非nan数值型', 则直接round()
    if is_notnan_numeric(x):
        ### 为了保证转成int, 需要先四舍五入, 再转成int
        ### 注意: 如果x=np.float(3.5000) , round(x, 0)  >>> 4.0 (还是会带个小数点,很烦,所以干脆转成int是最省心/最干净的)
        return int(round(x, 0))
    # 否则: 原样return回去
    else:
        if need_to_print:
            print(f"x: {x}, 类型为: {type(x)}, 不能保留整数!\n")
        return x







def round_df(df, round_digit=0, inplace=False, included_columns=[], excluded_columns=[]):
    """
        params:
            round_digit: 保留的小数位数
            inplace: 是否在原df上操作?
            excluded_columns: 排除某些不需要转化的列
            included_columns: 只有这些列 需要被转化


        默认:
            1. 四舍五入到整数
            2. 列名为"xx率"的, 一律以附带"百分号", 以str的形式呈现
    """
    if inplace is False:
        df = copy.deepcopy(df)

    for column, dtype in df.dtypes.items():
        # 1. 有些'率'是需要转化成'百分数'的
        if "率" in column:
            df[column] = df[column].apply(lambda x: format(x, ".2%"))
        # 2. 若发现df中的某列是 int型或者float型, 则按照round_digit四舍五入
        else:
            # 1. 当仅仅需要某几个列需要转化时:
            if included_columns:
                if column in included_columns:
                    if dtype == np.dtype(np.float64) or dtype == np.dtype(np.int64):
                        df[column] = df[column].round(round_digit)
            # 2. 当某几个列 一定不能转化时:
            elif excluded_columns:
                if column not in excluded_columns: # 只有'不被排除列'才需要保留两位小数
                    if dtype == np.dtype(np.float64) or dtype == np.dtype(np.int64):
                        df[column] = df[column].round(round_digit)
            # 3. 当都没有限制条件时:
            else:
                if dtype == np.dtype(np.float64) or dtype == np.dtype(np.int64):
                    df[column] = df[column].round(round_digit)
    return df



def get_random_num():
    "获取一个 [0.0, 1.0) 的随机数"
    return np.random.rand(1)[0]


def get_random_df(df):
    "对df的每一行打乱顺序"
    random_df = pd.DataFrame(np.random.permutation(df), columns=df.columns)
    return random_df



def md5(s: str):
    """
        function: 获取md5哈希编码后的值;
        返回类型: 为str
        notes: md5是不可逆的加密 (不属于对称加密和非对称加密)
    """
    MD5 = hashlib.md5()
    MD5.update(s.encode("utf-8"))
    encrypted_s = MD5.hexdigest()
    print(f"加密后的值为: {encrypted_s}\n")
    return encrypted_s


def create_encrypted_cookie(key: str, salt="618"):
    "通过加盐, 加时间数, 加随机数, 获得一个md5加密后的随机cookies (其实也没必要加密,只是用于记录登录状态,并没有其他作用)"
    s = key + salt + get_sim_this_time() + str(np.random.randint(10, 1000000))
    encrypted_s = md5(s)
    return encrypted_s





def kplot(df, kind="line"):
    """
        params:
            kind:
                line: 折线图
                bar: 条形图📊 (竖直型, 如左图标) [index上的索引, 就代表的是坐标轴上的标签]
                barh: 条形图 (水平型)
                hist: 直方图 (每个值的频率图) (类似于曝光直方图)
                pie: 饼状图

        todo:
            1. 如何对df中的某一列, 标注红色, 且加粗?? (其他设置成灰色/浅色虚线?)

    """
    import matplotlib.pyplot as plt
    plt.figure()
    df.plot()
    plt.show()








if __name__ == '__main__':
    print("start!")
    df = import_data("业务反馈调价", is_df=True)
    print(df)
    print(df.shape)
    print("end!")
